
local assert, pairs, ipairs, notice = assert, pairs, ipairs, notice
local math = math
local gl = require 'gl'; local GL = gl
local Node = require 'Node'
local Textures = require 'Textures'

module(...)

local Actor = _M
Node.Moving:subclass(Actor)

Actor.ix = nil
Actor.iy = nil
Actor.size = nil

Actor.alpha = 1
Actor.textures = nil
Actor.animationTime = nil
Actor.direction = 2

function Actor:initialize()
  self.visitedFields = self.visitedFields or {}
end

function Actor:loadTextures(info)
  self.textureInfo = info
  self.textures = { dlists = {}, delay = info.delay, frames = info.frames }
  for dir = 1, 8 do for frame = 0, info.frames do
    local tex = Textures.load(info.base..dir..frame..'.png{^2}')
    self.textures[dir..frame] = tex
    self.textures.dlists[dir..frame] = gl.GenLists(1)
    gl.NewList(self.textures.dlists[dir..frame], GL.COMPILE)
      gl.Translate(-(info.rel_x or 0), info.rel_y or 0, 0)
      gl.BindTexture(GL.TEXTURE_2D, tex.name)
      gl.Begin(GL.QUADS)
        gl.TexCoord(0, 0); gl.Vertex(0, 0)
        gl.TexCoord(tex.wratio, 0); gl.Vertex(tex.origw, 0)
        gl.TexCoord(tex.wratio, tex.hratio); gl.Vertex(tex.origw, tex.origh)
        gl.TexCoord(0, tex.hratio); gl.Vertex(0, tex.origh)
      gl.End()
    gl.EndList()
  end end
end

function Actor:displayContent()
  --[=[
  gl.BindTexture(GL.TEXTURE_2D, 0)
  local tw, th = self.parent.data.tilewidth, self.parent.data.tileheight
  gl.Begin(GL.QUADS)
    gl.Color(1, 0, 0); gl.Vertex(tw*self.size/2, 0)
    gl.Color(1, 1, 0); gl.Vertex(tw*self.size, th*self.size/2)
    gl.Color(0, 1, 0); gl.Vertex(tw*self.size/2, th*self.size)
    gl.Color(0, 0, 1); gl.Vertex(0, th*self.size/2)
  gl.End()
  gl.Color(1, 1, 1)
  --]=]
  local frame = 0
  if self.animationTime then
    frame = self.animationTime + self:getLoop().frameFraction
    frame = math.floor(frame/self.textures.delay)
    frame = (frame % self.textures.frames) + 1
  end
  gl.Color(self.red or 1, self.green or 1, self.blue or 1, self.alpha)
  gl.CallList(self.textures.dlists[self.direction..frame])
  gl.Color(1, 1, 1)
end

function Actor:update()
  if self.fade then
    self.alpha = math.max(0, math.min(1, self.alpha + self.fade))
    if self.alpha == 0 or self.alpha == 1 then self.fade = nil end
    if self.dying then self.parent:removeChild(self) end
  end
  if self.throwing then
    self.throwing = self.throwing - 1
    self:setPosition(self:collide(self.ix, self.iy,
      self.throwDirection, self.throwSpeed))
    if self.throwing <= 0 then self.throwing = nil end
  end
end

function squaresIntersect(x1, y1, s1, x2, y2, s2)
  x1 = math.floor(x1*10000+0.5); x2 = math.floor(x2*10000+0.5)
  y1 = math.floor(y1*10000+0.5); y2 = math.floor(y2*10000+0.5)
  s1 = math.floor(s1*10000+0.5); s2 = math.floor(s2*10000+0.5)
  return
    math.min(x1+s1, x2+s2) > math.max(x1, x2)+10 and
    math.min(y1+s1, y2+s2) > math.max(y1, y2)+10
end

function Actor:intersectsField(x, y, myx, myy)
  return squaresIntersect(x, y, 1, myx or self.ix, myy or self.iy, self.size)
end

function Actor:intersectedFields(selfx, selfy)
  local result = {}
  selfx, selfy = selfx or self.ix, selfy or self.iy
  for y = math.floor(selfy), math.ceil(selfy+self.size) do
    for x = math.floor(selfx), math.ceil(selfx+self.size) do
      if self:intersectsField(x, y, selfx, selfy) then
        result[#result+1] = { x, y }
      end
    end
  end
  return result
end

function Actor:lastIntersectedField()
  return math.floor(self.ix+self.size-0.1), math.floor(self.iy+self.size-0.1)
end

function Actor:movedInto(x, y)
  self.visitedFields[#self.visitedFields+1] = { x, y }
end

function Actor:setPosition(ix, iy)
  local oldx, oldy = self.ix, self.iy
  self.ix, self.iy = ix, iy
  -- find the orthogonal lower left corner of the square
  self.x = math.floor((self.ix - (self.iy+self.size) + 1)/2 *
    self.parent.data.tilewidth + 0.5)
  self.y = math.floor((self.ix+self.size + self.iy+self.size - 2)/2 *
    -self.parent.data.tileheight + 0.5)
  local z = 0
  for i,v in ipairs(self:intersectedFields()) do
    for i = 1, #self.parent.data.layers do
      local tileinfo = self.parent.cache[i][v[2]] and
        self.parent.cache[i][v[2]][v[1]]
      z = math.max(z, tileinfo and tileinfo.height or 0)
    end
  end
  self.y = self.y + z
  --self.parent:updateActorPositions()
  -- handle movedInfo
  for i,v in ipairs(self:intersectedFields()) do
    if not self:intersectsField(v[1], v[2], oldx, oldy) then
      self:movedInto(v[1], v[2])
    end
  end
end

function Actor:canAccessField(ix, iy)
  for i = 1, #self.parent.data.layers do
    local tileinfo = self.parent.cache[i][iy] and self.parent.cache[i][iy][ix]
    if tileinfo and tileinfo.noaccess then return false end
  end
  return true
end

function Actor:goBackIfStuck()
  local good = {}; local stuck = false
  for i,v in ipairs(self:intersectedFields()) do
    if self:canAccessField(v[1], v[2]) then
      good[#good+1] = v
    else
      stuck = true
    end
  end
  if not stuck then return end
  if good[1] then return self:setPosition(good[1][1], good[1][2]) end
  for i = #self.visitedFields, 1, -1 do
    local x, y = self.visitedFields[i][1], self.visitedFields[i][2]
    if self:canAccessField(x, y) then return self:setPosition(x, y) end
  end
  -- still stuck? too bad.
end

function Actor:collidesWith(actor)
  if self.cantCollide or self.ghost then return false end
  -- allow moving when already intersecting another actor (e.g. because of
  -- having been teleported there)
  if squaresIntersect(self.ix, self.iy, self.size,
    actor.ix, actor.iy, actor.size) then return false end
  return true
end

function movePointTowardsSquare(px, py, dx, dy, sx, sy, ss)
  -- begone, annoying rounding errors! HA!
  px = math.floor(px*10000+0.5); py = math.floor(py*10000+0.5)
  dx = math.floor(dx*10000+0.5); dy = math.floor(dy*10000+0.5)
  sx = math.floor(sx*10000+0.5); sy = math.floor(sy*10000+0.5)
  ss = math.floor(ss*10000+0.5)
  local nearestT = 1; local t
  -- left edge (px+t*dx = sx)
  t = (sx-px)/dx
  if py+t*dy >= sy and py+t*dy <= sy+ss and t >= 0 then
    nearestT = math.min(nearestT, t) end
  -- right edge (px+t*dx = sx+ss)
  t = (sx+ss-px)/dx
  if py+t*dy >= sy and py+t*dy <= sy+ss and t >= 0 then
    nearestT = math.min(nearestT, t) end
  -- top edge (py+t*dy = sy)
  t = (sy-py)/dy
  if px+t*dx >= sx and px+t*dx <= sx+ss and t >= 0 then
    nearestT = math.min(nearestT, t) end
  -- bottom edge (py+t*dy = sy+ss)
  t = (sy+ss-py)/dy
  if px+t*dx >= sx and px+t*dx <= sx+ss and t >= 0 then
    nearestT = math.min(nearestT, t) end
  -- special case: don't get stuck on an edge when moving away
  if nearestT == 0 then
    if (px == sx and py >= sy and py <= sy+ss and dx <= 0) or       -- left
      (px == sx+ss and py >= sy and py <= sy+ss and dx >= 0) or     -- right
      (py == sy and px >= sx and px <= sx+ss and dy <= 0) or        -- top
      (py == sy+ss and px >= sx and px <= sx+ss and dy >= 0) then   -- bottom
--        notice'specialcase'
        return 1 end
  end
  -- more rounding error prevention
  return nearestT == 1 and 1 or math.min(nearestT-0.001, 0)
end

function Actor:findCollision(myx, myy, dx, dy)
  local moveDistance = 1
  local function collisionAtDistance(d)
--    notice('md', moveDistance, d)
    moveDistance = math.min(moveDistance, d)
  end
  local function collideWithSquare(x, y, s)
    -- send my verticles into the square
    -- (also check points at edge centers so that two squares of the same size
    -- can't intersect because of the movePointTowardsSquare special case)
    for ux = 0, 1, 0.5 do for uy = 0, 1, 0.5 do
      -- send my verticles into the square
--      notice('shoot', ux, uy, myx+self.size*ux, myy+self.size*uy, dx, dy, x, y, s)
      collisionAtDistance(movePointTowardsSquare(
        myx+self.size*ux, myy+self.size*uy, dx, dy, x, y, s))
      -- send the square's verticles into myself
--      notice('shoot', ux, uy, x+s*ux, y+s*uy, -dx, -dy, myx, myy, self.size)
      collisionAtDistance(movePointTowardsSquare(
        x+s*ux, y+s*uy, -dx, -dy, myx, myy, self.size))
    end end
  end
  -- check nearby tiles for accessibility
  for x = math.floor(myx+math.min(dx, 0)),
      math.ceil(myx+self.size+math.max(dx, 0)) do
    for y = math.floor(myy+math.min(dy, 0)),
        math.ceil(myy+self.size+math.max(dy, 0)) do
      if not self:canAccessField(x, y) then collideWithSquare(x, y, 1) end
    end
  end
  -- check actors for collisions
  for i,v in ipairs(self.parent:children()) do
    if v ~= self and self:collidesWith(v) and v:collidesWith(self) then
      collideWithSquare(v.ix, v.iy, v.size)
    end
  end
  return moveDistance
end

function Actor:findCollisionQuickly(myx, myy, dx, dy)
  -- TODO don't require this function... it's soo ugly
  local moveDistance = 1
  local function collisionAtDistance(d)
--    notice('md', moveDistance, d)
    moveDistance = math.min(moveDistance, d)
  end
  local function collideWithSquare(x, y, s)
    if squaresIntersect(x, y, s, myx, myy, self.size) or
        squaresIntersect(x, y, s, myx+dx, myy+dy, self.size) then
      collisionAtDistance(0)
    end
  end
  -- check nearby tiles for accessibility
  for x = math.floor(myx+math.min(dx, 0)),
      math.ceil(myx+self.size+math.max(dx, 0)) do
    for y = math.floor(myy+math.min(dy, 0)),
        math.ceil(myy+self.size+math.max(dy, 0)) do
      if not self:canAccessField(x, y) then collideWithSquare(x, y, 1) end
    end
  end
  -- check actors for collisions
  for i = self.sortedId-3, self.sortedId+3 do
    local v = self.parent.sortedActors[i]
    if v and v ~= self and self:collidesWith(v) and v:collidesWith(self) then
      collideWithSquare(v.ix, v.iy, v.size)
    end
  end
  return moveDistance
end

function Actor:collide(myx, myy, direction, distance, recursed)
  -- TODO test the recent modifications
  local dx, dy = self.parent:getDirection(direction, distance)
  local moveDistance = self:findCollision(myx, myy, dx, dy)
  local newx, newy = myx+moveDistance*dx, myy+moveDistance*dy
--  notice('result:', recursed, myx, myy, dx, dy, '->', moveDistance, newx, newy)
  if moveDistance == 1 or recursed then return newx, newy, moveDistance end
  -- try nearby directions
  local t1x, t1y, t1d = self:collide(newx, newy,
    (direction-2) % 8 + 1, distance*(1-moveDistance)/math.sqrt(2), true)
  local t2x, t2y, t2d = self:collide(newx, newy,
    direction % 8 + 1, distance*(1-moveDistance)/math.sqrt(2), true)
  if t1d == 0 and t2d == 0 then return newx, newy, moveDistance end
  if t2d > t1d then return t2x, t2y, t2d else return t1x, t1y, t1d end
end

function Actor:move(direction, distance)
  self.direction = direction
  self.animationTime = (self.animationTime or 0) + 1
  if distance == 0 then self.animationTime = nil; return end
  if self.cantCollide then
    local dx, dy = self.parent:getDirection(direction, distance)
    self:setPosition(self.ix+dx, self.iy+dy)
    return 1
  else
    local nx, ny, dist = self:collide(self.ix, self.iy, direction, distance)
    self:setPosition(nx, ny)
    return dist
  end
end

function Actor:throwBack(direction, speed, time)
  self.throwing = time
  self.throwSpeed = speed
  self.throwDirection = direction
end

