
local assert, pairs, ipairs, notice = assert, pairs, ipairs, notice
local math = math
local Textures = require 'Textures'
local gl = require 'gl'; local GL = gl
local Actor = require 'yarivan.Actor'

module(...)

local Enemy = _M
Actor:subclass(Enemy)

Enemy.isEnemy = true

Enemy.health = 2
Enemy.speed = 2.6
Enemy.pursuitSpeed = 3.2
Enemy.startPursuitDistance = 5
Enemy.stopPursuitDistance = 9
Enemy.walkTimeMin = 10
Enemy.walkTimeMax = 70
Enemy.waitTimeMin = 0
Enemy.waitTimeMax = 10

Enemy.walkTime = 0
Enemy.waitTime = 0
Enemy.direction = 1
Enemy.pursuing = nil

function Enemy:loadTextures(info)
  -- TODO avoid needless code duplication with Actor:loadTextures()
  self.textureInfo = info
  self.textures = { dlists = {}, delay = info.delay, frames = info.frames }
  for dir = 1, 8, 2 do for frame = 1, info.frames do
    local tex = Textures.load(info.base..dir..frame..'.png{^2}')
    self.textures[dir..frame] = tex
    self.textures.dlists[dir..frame] = gl.GenLists(1)
    if frame == 1 then
      self.textures[dir..'0'] = tex
      self.textures.dlists[dir..'0'] = self.textures.dlists[dir..frame]
    end
    gl.NewList(self.textures.dlists[dir..frame], GL.COMPILE)
      gl.Translate(-(info.rel_x or 0), info.rel_y or 0, 0)
      gl.BindTexture(GL.TEXTURE_2D, tex.name)
      gl.Begin(GL.QUADS)
        gl.TexCoord(0, 0); gl.Vertex(0, 0)
        gl.TexCoord(tex.wratio, 0); gl.Vertex(tex.origw, 0)
        gl.TexCoord(tex.wratio, tex.hratio); gl.Vertex(tex.origw, tex.origh)
        gl.TexCoord(0, tex.hratio); gl.Vertex(0, tex.origh)
      gl.End()
    gl.EndList()
  end end
end

function Enemy:update()
  Actor.update(self)
  if self.damaged then
    self.damaged = self.damaged - 1
    if self.damaged > 0 then return end
    self.damaged, self.red, self.green, self.blue = nil
    if self.health > 0 then return end
    self.parent:removeChild(self)
    return
  end
  -- check whether to pursue a player
  local world = self.parent.parent.parent.parent
  local playerDists = {}
  local newPursuit = nil
  local start = self.startPursuitDistance*self.startPursuitDistance
  local stop = self.stopPursuitDistance*self.stopPursuitDistance
  local myX, myY = self.ix+self.size/2, self.iy+self.size/2
  local bestX, bestY, bestDist = nil
  for i,v in ipairs(world.players) do
    local pX, pY = v.actor.ix+v.actor.size/2, v.actor.iy+v.actor.size/2
    local dx, dy = pX-myX, pY-myY; local length2 = dx*dx+dy*dy
    if v.map == self.parent and not v.actor.ghost and length2 < (bestDist or math.huge) and
        length2 < (self.pursuing == i and stop or start) then
      self.pursuing = i; bestX, bestY, bestDist = dx, dy, length2
    elseif self.pursuing == i then
      self.pursuing = nil
    end
  end
  if self.pursuing then
    self.walkTime = 0; self.waitTime = 0
    local realPursuitSpeed = self.pursuitSpeed * self.parent.directions[1][2]
    local speedFactor = realPursuitSpeed / math.sqrt(bestDist)
    self:moveBy(bestX*speedFactor, bestY*speedFactor)
    local threshold = 0.2
    local actor = world.players[self.pursuing].actor
    if Actor.squaresIntersect(actor.ix, actor.iy, actor.size,
        self.ix-threshold, self.iy-threshold, self.size+2*threshold) then
      world.players[self.pursuing]:damage()
      world.players[self.pursuing]:throwBack(self)
    end
    return
  end
  -- walk around randomly instead
  if self.walkTime <= 0 then
    if self.waitTime > 0 then
      self.waitTime = self.waitTime - 1
      self:move(self.direction, 0); return
    end
    self.direction = ( (self.direction - 2 + math.random(2)*4) - 1) % 8 + 1
    self.walkTime = math.random(self.walkTimeMin, self.walkTimeMax)
    self.waitTime = math.random(self.waitTimeMin, self.waitTimeMax)
  end
  self.walkTime = self.walkTime - 1
  local distance = self:move(self.direction, self.speed)
  if distance < 1 then self.walkTime = 0 end   -- change direction
end

function Enemy:move(direction, distance)
  self.direction = direction
  self.animationTime = (self.animationTime or 0) + 1
  if distance == 0 then self.animationTime = nil; return end
  local dx, dy = self.parent:getDirection(direction, distance)
  local moveDistance = self:findCollisionQuickly(self.ix, self.iy, dx, dy)
  self:setPosition(self.ix+moveDistance*dx, self.iy+moveDistance*dy)
  return moveDistance
end

function Enemy:moveBy(dx, dy)
  local theta = math.atan2(-dy, dx)
  if theta < -3*math.pi/4 then self.direction = 7
  elseif theta < -math.pi/4 then self.direction = 1
  elseif theta < math.pi/4 then self.direction = 3
  elseif theta < 3*math.pi/4 then self.direction = 5
  else self.direction = 7 end
  self.animationTime = (self.animationTime or 0) + 1
  local moveDistance = self:findCollision(self.ix, self.iy, dx, dy)
  self:setPosition(self.ix+moveDistance*dx, self.iy+moveDistance*dy)
  if moveDistance < 1 then
    dx = dx * (1-moveDistance); dy = dy * (1-moveDistance)
    local xDistance = self:findCollision(self.ix, self.iy, dx, 0)
    local yDistance = self:findCollision(self.ix, self.iy, 0, dy)
    if math.abs(xDistance*dx) > math.abs(yDistance*dy) then
      self:setPosition(self.ix+xDistance*dx, self.iy)
    elseif yDistance > 0 and dy ~= 0 then
      self:setPosition(self.ix, self.iy+yDistance*dy)
    end
  end
end

function Enemy:damage(points)
  self.health = self.health - (points or 1)
  self.red, self.green, self.blue = (self.health <= 0 and 0 or 1), 0, 0
  self.damaged = 1
end

