
local assert, pairs, ipairs, notice = assert, pairs, ipairs, notice
local setmetatable, tonumber = setmetatable, tonumber
local string = string
local SDL = require 'SDL'
local Node = require 'Node'
local Event = require 'Event'
local Map = require 'yarivan.Map'
local MapView = require 'yarivan.MapView'
local Player = require 'yarivan.Player'
local Music = require 'yarivan.Music'
local FS = FS
require 'extlib'; local inspect = inspect
local _G = _G   -- TODO move WIDTH and HEIGHT elsewhere

module(...)

local GameWorld = _M
Node:subclass(GameWorld)

GameWorld.time = 0
GameWorld.mapviews = nil
GameWorld.players = nil

function GameWorld:getMap(mapfile)
  for i,v in ipairs(self.players) do
    if v.map and v.map.mapfile == mapfile then return v.map end
  end
  return nil
end

function GameWorld:loadedMaps()
  local maps = {}
  for i,v in ipairs(self.players) do if v.map then maps[v.map] = true end end
  return maps
end

function GameWorld:initialize()
  if not self.players then
    -- TODO not hardcoded controls
    self:addChild('players', Node:new{ parent = self })
    self.players[1] = Player:new{ parent = self.players, id = 1,
      key_left = SDL.K_a, key_right = SDL.K_d, key_up = SDL.K_w,
      key_down = SDL.K_s, key_activate = SDL.K_SPACE,
      key_shoot = SDL.K_f, actorTextures = { base = 'jaguar/',
      rel_x = 0, rel_y = 10, frames = 4, delay = 2.8 } }
    self.players[2] = Player:new{ parent = self.players, id = 2,
      key_left = SDL.K_LEFT, key_right = SDL.K_RIGHT, key_up = SDL.K_UP,
      key_down = SDL.K_DOWN, key_activate = SDL.K_RETURN,
      key_shoot = SDL.K_RCTRL, actorTextures = { base = 'platypus/',
      rel_x = 0, rel_y = 10, frames = 4, delay = 2.5 } }
  end
  if not self.mapviews then
    self:addChild('mapviews', Node:new{ parent = self })
    for i,v in ipairs({ 0, _G.WIDTH/2+2 }) do
      self.mapviews[i] = MapView:new{ x = v, y = 0, w = _G.WIDTH/2-2,
        h = _G.HEIGHT, player = self.players[i], parent = self.mapviews }
    end
  end
  self.state = self.state or {}
end

local musiclist

function GameWorld:update()
  self.time = self.time + 1
  if Event.key[SDL.K_ESCAPE] then
    self:getLoop():subLoop(_G.MainLoop, { root = _G.require'yarivan.MainMenu'.makeGameMenu() })   -- TODO soo ugly
    return
  end
  if self.frozen then
    if self.frozen > 0 then self.frozen = self.frozen - 5 end
  else
    self.players[1]:processInput()
    self.players[2]:processInput()
  end
  for map,_ in pairs(self:loadedMaps()) do map:update() end
  Node.update(self)
  if not musiclist then musiclist = {}; for k,v in pairs(FS.listFiles('music/')) do musiclist[#musiclist+1] = 'music/'..k end end
  Music.update(musiclist)
end

function GameWorld:updateState()
  for map,_ in pairs(self:loadedMaps()) do map:updateState() end
end

function GameWorld:saveGame(id, name)
  -- TODO de-uglify it
  -- TODO persistent cache and animTimes
  local toRemove = {}
  for fn,ft in pairs(FS.listFiles('save/')) do
    if fn:match('^'..id..'%-') then toRemove[#toRemove+1] = fn end
  end
  for i,v in ipairs(toRemove) do FS.remove('save/'..v) end
  local body = [[require 'Node'
require 'yarivan.WindowManager'
require 'yarivan.GameWorld'
local root = Node.Root:new{}
root:addChild('wm', yarivan.WindowManager:new{ parent = root })
root.wm:addChild('world', yarivan.GameWorld:new]]..inspect.table(self, inspect.export,
  inspect.whitelist({ frozen = true, time = true, state = true })):gsub('}$', ', parent = root.wm}')..[[);]]
  for i,v in ipairs(self.players) do
    body = body .. [[
local p = root.wm.world.players[]]..i..[[]
p:setLocation(]]..inspect.object(v.map.mapfile)..[[, ]]..v.actor.ix..[[, ]]..v.actor.iy..[[)
for k,v in pairs(]]..inspect.table(v, inspect.export,
  inspect.blacklist({ map=true,parent=true,actor=true,warping=true }))..[[) do p[k] = v end;
for k,v in pairs(]]..inspect.table(v.actor, inspect.export,
  inspect.blacklist({ parent=true,player=true,textures=true }))..[[) do p.actor[k] = v end
if not (p.map[1] and p.map[1].player) then
for i = 1, #p.map do
  if p.map[1].player then break end; table.remove(p.map, 1) end
]]
    for _,vv in ipairs(v.map) do if vv.isEnemy then
      body = body .. [[
p.map:addEnemy(]]..vv.ix..[[, ]]..vv.iy..[[, ]]..inspect.table(vv, inspect.export,
  inspect.blacklist{ parent=true, textureInfo=true, textures=true })..[[, ]]..inspect.table(vv.textureInfo, inspect.export)..[[)
      ]]
    end end
    body = body .. [[ end ]]
  end
  body = body .. [[return root]]
  FS.write('save/'..id..'-'..encodeSavegameName(name)..'.lua', body)
  _G.USERRC = _G.USERRC..'\nLASTGAME = '.._G.LASTGAME
  _G.pcall(FS.write, 'userrc.lua', _G.USERRC)
end

function encodeSavegameName(s)
  return s:gsub('[^ !#-)+%-0-9;=@-[%]^-{}~]',
    function (ch) return (',%02X'):format(ch:byte()) end)
end

function decodeSavegameName(s)
  return s:gsub(',(%x%x)',
    function (h) return string.char(tonumber(h, 16)) end)
end

