
local os = os
local assert, pairs, ipairs, notice = assert, pairs, ipairs, notice
local tonumber = tonumber
local Node = require 'Node'
local MainLoop = require 'MainLoop'
local Menu = require 'yarivan.Menu'
local WindowManager = require 'yarivan.WindowManager'
local GameWorld = require 'yarivan.GameWorld'
local Textures = require 'Textures'
local FS = FS
local _G = _G

module(...)

maxSavedGames = 10

function makeMainMenu(o)
  o = o or {}
  o.background = Textures.load'menu.png{^2}'
  o.items = { "Nová hra", "Načítať hru", "Nastavenia", "Koniec hry" }
  o.hooks = {
    function (self)
      self.fade = -0.2
      self.finish = function ()
        self:getLoop():replaceLoop(MainLoop, { root = makeNewGame() })
      end
    end,
    function (self)
      self.fade = -0.2
      self.finish = function ()
        self:getLoop():replaceLoop(MainLoop, { root = makeLoadMenu() })
      end
    end,
    function (self)
      self.fade = -0.2
      self.finish = function ()
        self:getLoop():replaceLoop(MainLoop, { root = makeSettingsMenu() })
      end
    end,
    function (self)
      self.fade = -0.2
    end,
  }
  return Menu:new(o)
end

function makeGameMenu(o)
  o = o or {}
  o.escape = 1
  o.items = { "Späť do hry", "Uložiť hru", "Načítať hru", "Nastavenia", "Koniec hry" }
  o.hooks = {
    function (self)
      self.fade = -0.2
    end,
    function (self)
      self.fade = -0.2
      self.finish = function ()
        self:getLoop():replaceLoop(MainLoop, { root = makeSaveMenu() })
      end
    end,
    function (self)
      self.fade = -0.2
      self.finish = function ()
        self:getLoop():replaceLoop(MainLoop, { root = makeLoadMenu(nil, true) })
      end
    end,
    function (self)
      self.fade = -0.2
      self.finish = function ()
        self:getLoop():replaceLoop(MainLoop, { root = makeSettingsMenu(nil, true) })
      end
    end,
    function (self)
      self.fade = -0.2
      self.finish = function ()
        self:getLoop().running = false
        self:getLoop().parentloop.running = false
      end
    end,
  }
  return Menu:new(o)
end

function makeLoadMenu(o, ingame)
  o = o or {}
  if not ingame then o.background = Textures.load'menu.png{^2}' end
  o.items = {}
  o.hooks = {}
  o.focusTo = _G.LASTGAME or 1
  local savedgames = {}
  local function loadgame(self, i)
    if not savedgames[i] then return end
    self.fade = -0.2
    self.finish = function ()
      _G.LASTGAME = i
      local fname = 'save/'..savedgames[i]
      local loop = self:getLoop()
      loop.running = false
      if loop.parentloop then loop = loop.parentloop end
      loop:replaceLoop(MainLoop, { root = assert(FS.loadCode(fname))() })
    end
  end
  for i = 1, maxSavedGames do
    o.items[i] = i..". Prázdna pozícia"; o.hooks[i] = loadgame
  end
  local saves = FS.listFiles('save/')
  for fname,ftype in pairs(saves) do
    local id, name = fname:match('^(%d+)-(.*).lua$')
    id = tonumber(id)
    savedgames[id] = fname
    o.items[id] = id..". "..GameWorld.decodeSavegameName(name)
  end
  o.items[#o.items+1] = "Späť"
  o.hooks[#o.hooks+1] = function (self)
    self.fade = -0.2
    self.finish = function ()
      if ingame then
        self:getLoop():replaceLoop(MainLoop, { root = makeGameMenu({ focusTo = 3 }) })
      else
        self:getLoop():replaceLoop(MainLoop, { root = makeMainMenu({ focusTo = 2 }) })
      end
    end
  end
  o.escape = #o.items
  return Menu:new(o)
end

function makeSaveMenu(o)
  o = o or {}
  o.items = {}
  o.hooks = {}
  o.focusTo = _G.LASTGAME or 1
  local savedgames = {}
  local function savegame(self, i)
    self.fade = -0.2
    self.finish = function ()
      _G.LASTGAME = i
      self:getLoop().parentloop.root.wm.world:saveGame(i, os.date('%Y-%m-%d %H:%M'))
      self:getLoop().running = false
    end
  end
  for i = 1, maxSavedGames do
    o.items[i] = i..". Prázdna pozícia"; o.hooks[i] = savegame
  end
  local saves = FS.listFiles('save/')
  for fname,ftype in pairs(saves) do
    local id, name = fname:match('^(%d+)-(.*).lua$')
    id = tonumber(id)
    savedgames[id] = fname
    o.items[id] = id..". "..GameWorld.decodeSavegameName(name)
  end
  o.items[#o.items+1] = "Späť"
  o.hooks[#o.hooks+1] = function (self)
    self.fade = -0.2
    self.finish = function ()
      self:getLoop():replaceLoop(MainLoop, { root = makeGameMenu({ focusTo = 2 }) })
    end
  end
  o.escape = #o.items
  return Menu:new(o)
end

function makeNewGame()
  local root = Node.Root:new{}
  root:addChild('wm', WindowManager:new{ parent = root })
  root.wm:addChild('world', GameWorld:new{ parent = root.wm })
  root.wm.world.players[1]:setLocation('maps/level0.tmx', 45, 59)
  root.wm.world.players[2]:setLocation('maps/level0.tmx', 72, 46)
  return root
end

function makeSettingsMenu(o, ingame)
  local haveSound = false
  for k,v in pairs(FS.listFiles('/music')) do haveSound = true; break end
  if not _G._SOUND then haveSound = false end
  o = o or {}
  if not ingame then   o.background = Textures.load'menu.png{^2}' end
  o.escape = 3
  o.items = {
    "Celá obrazovka: "..(_G.FULLSCREEN and "ZAP" or "VYP"),
    "Hudba: "..(haveSound and (_G.SOUND and "ZAP" or "VYP") or "N/A"),
    "Späť"
  }
  o.hooks = {
    function (self)
      _G.FULLSCREEN = not _G.FULLSCREEN
      _G.USERRC = _G.USERRC..'\nFULLSCREEN = '..(_G.FULLSCREEN and 'true' or 'false')
      _G.pcall(FS.remove, 'userrrc.lua')
      _G.pcall(FS.write, 'userrc.lua', _G.USERRC)
      self.fade = -0.2
      local restartMessage = Menu:new{
        escape = 1,
        items = { "Zmeny sa prejavia po reštarte hry." },
        hooks = { function (subself) subself.fade = -0.2
          subself.finish = function () subself:getLoop():replaceLoop(MainLoop,
            { root = makeSettingsMenu() }) end end }
      }
      self.finish = function ()
        self:getLoop():replaceLoop(MainLoop, { root = restartMessage })
      end
    end,
    function (self)
      if not haveSound then return end
      _G.SOUND = not _G.SOUND
      if not _G.SOUND then _G.yarivan.Music.stop() end
      _G.USERRC = _G.USERRC..'\nSOUND = '..(_G.SOUND and 'true' or 'false')
      _G.pcall(FS.remove, 'userrrc.lua')
      _G.pcall(FS.write, 'userrc.lua', _G.USERRC)
      self.items[2] = "Hudba: "..(_G.SOUND and "ZAP" or "VYP")
    end,
    function (self)
      self.fade = -0.2
      self.finish = function ()
        if ingame then
          self:getLoop():replaceLoop(MainLoop, { root = makeGameMenu({ focusTo = 4 }) })
        else
          self:getLoop():replaceLoop(MainLoop, { root = makeMainMenu({ focusTo = 3 }) })
        end
      end
    end
  }
  return Menu:new(o)
end

