
local Node = require 'Node'
local MapLoader = require 'yarivan.MapLoader'
local Enemy = require 'yarivan.Enemy'
local assert, pairs, ipairs, notice = assert, pairs, ipairs, notice
local setfenv, tonumber, loadstring, setmetatable =
  setfenv, tonumber, loadstring, setmetatable
local math, table = math, table
local _G = _G

module(...)

local Map = _M
Node:subclass(Map)

function Map:initialize()
  self.initTime = self.parent.parent.parent.time
  self.doorcolor = self.doorcolor or {}
  if not self.data then
    self.data = MapLoader.load(assert(self.mapfile, 'nil mapfile'))
    self.hooks = {}
    self.cache = {}
    self.animTimes = {}
    for i = 1, #self.data.layers do
      self.animTimes[i] = {}
      self.cache[i] = {}
      for y,row in pairs(self.data.layers[i].absfields) do
        self.animTimes[i][y] = {}
        self.cache[i][y] = {}
        for x,gid in pairs(row) do
          self.cache[i][y][x] = self.data.tiles[gid]
        end
      end
    end
    local script = assert(loadstring(self.data.script, '@'..self.mapfile))
    local env = setmetatable({ map = self,
      world = self.parent.parent.parent }, { __index = _G })
    function env.hook(tile)
      tile = self:parseTile(tile)
      self.hooks[tile] = self.hooks[tile] or {}
      return self.hooks[tile]
    end
    env._S = env; setfenv(script, env); script()
  end
  if not self.directions then
    -- all direction vectors should have the same length on-screen (in
    -- scaled orthometric view)
    local tw, th = self.data.tilewidth, self.data.tileheight
    local diag = math.sqrt(4/(tw*tw+th*th))
    self.directions = {
      {0, diag},
      {1/th, 1/th},
      {diag, 0},
      {1/tw, -1/tw},
      {0, -diag},
      {-1/th, -1/th},
      {-diag, 0},
      {-1/tw, 1/tw}
    }
    self.directionAngles = {
      {0, 1},
      {1, 1},
      {1, 0},
      {1, -1},
      {0, -1},
      {-1, -1},
      {-1, 0},
      {-1, 1},
    }
  end
end

function Map:getDirection(dir, distance)
  return distance*self.directions[dir][1], distance*self.directions[dir][2]
end

function Map:parseTile(tile)
  local tileset, offset = (''..tile):match('^(.*)/(%d+)$')
  if tileset then return tonumber(offset) +
    assert(self.data.firstgids[tileset], 'no such tileset') end
  return tonumber(tile)
end

function Map:find(tile)
  -- results sorted by ascending i, ascending y, ascending x.
  tile = self:parseTile(tile)
  local results = {}
  for i = 1, #self.data.layers do
    local layer = self.data.layers[i]
    for y = 0, #layer.fields do for x = 0, #layer.fields[y] do
      if layer.fields[y][x] == tile then
        results[#results+1] = { i = i, x = x+layer.x, y = y+layer.y }
      end
    end end
  end
  return results
end

function Map:setField(tile, layerid, x, y)
  if not x then layerid, x, y = layerid.i, layerid.x, layerid.y end
  if not y then x, y = x.x, x.y end
  tile = self:parseTile(tile)
  local old = self.cache[layerid][y][x]
  self.cache[layerid][y][x] = self.data.tiles[tile]
  if self.data.tiles[tile].noloop and old ~= self.data.tiles[tile] then
    self.animTimes[layerid][y][x] = self.parent.parent.parent.time end
  for _,actor in ipairs(self:children()) do
    actor:goBackIfStuck()
    actor:setPosition(actor.ix, actor.iy)   -- update tile height
  end
end

function Map:addEnemy(x, y, enemy, textures)
  enemy.parent = self
  enemy = Enemy:new(enemy)
  enemy:loadTextures(textures)
  self[#self+1] = enemy
  enemy:setPosition(x, y)
end

function Map:updateState()
end

function Map:updateActorPositions()
  self.sortedActors = self:children()
  table.sort(self.sortedActors, function (a,b)
    return (a.ix+a.size + a.iy+a.size) < (b.ix+b.size + b.iy+b.size) end)
  for i,v in ipairs(self.sortedActors) do v.sortedId = i end
end

function Map:callHook(x, y, hook, ...)
  --notice('callHook '..hook, x, y)
  local fired = false
  local world = self.parent.parent.parent
  for layerid = 1, #self.data.layers do
    local layer = self.data.layers[layerid]
    local hookinfo = self.hooks[layer.absfields[y] and layer.absfields[y][x]]
    if hookinfo and hookinfo[hook] then
      fired = true; hookinfo[hook](x, y, layerid, ...)
    end
  end
  if fired then world:updateState() end
end

function Map:update()
  Node.update(self)
  self:updateActorPositions()
  if self.parent.parent.parent.time % 50 == 0 then
    -- protection against getting stuck
    for _,actor in ipairs(self:children()) do
      actor:goBackIfStuck()
      actor:setPosition(actor.ix, actor.iy)   -- update tile height
    end
  end
  -- TODO respawn enemies
end

function Map:displayContent()
  -- that's what MapView does
end

