
require 'buf'; local string = string
local bit = require 'bit'
local SDL = require 'SDL'
local gl = require 'gl'; local GL = gl
local zlib = require 'zlib'
local FS = FS
local table = table
local assert, pairs, ipairs, notice = assert, pairs, ipairs, notice
local tonumber, setmetatable = tonumber, setmetatable
local Textures = require 'Textures'
local _G = _G

module(...)

local decode_t = {}
for i = 1, 26 do decode_t[i + 64] = i - 1 end
for i = 1, 26 do decode_t[i + 96] = i - 1 + 26 end
for i = 0, 9 do decode_t[i + 48] = i + 26 + 26 end
decode_t[43] = 26 + 26 + 10
decode_t[47] = 26 + 26 + 10 + 1
decode_t[61] = 0

local function base64_decode(s)
  assert(#s%4 == 0, "malformed base64")
  local out = {}
  for i = 0, #s-1, 4 do
    local m = bit.bor(
      bit.lshift(decode_t[string.byte(s, i+1, i+1)], 18),
      bit.lshift(decode_t[string.byte(s, i+2, i+2)], 12),
      bit.lshift(decode_t[string.byte(s, i+3, i+3)], 6),
      bit.lshift(decode_t[string.byte(s, i+4, i+4)], 0))
    table.insert(out, string.char(
      bit.band(bit.rshift(m, 16), 0xFF),
      bit.band(bit.rshift(m, 8), 0xFF),
      bit.band(m, 0xFF)))
  end
  if s:sub(-2) == '==' then out[#out] = out[#out]:sub(1, 1)
  elseif s:sub(-1) == '=' then out[#out] = out[#out]:sub(1, 2)
  end
  return table.concat(out)
end

local function xmlChildren(element)
  local ret = {}
  for i,v in ipairs(element) do if v.xml then ret[v.xml] = v end end
  return ret
end

local function loadProperties(table, propertiesElement, namespace)
  table = table or {}
  for i,v in ipairs(propertiesElement or {}) do table[v.name] = v.value end
  return table
end

local mapCache = setmetatable({}, { __mode = 'v' })

local tilesetDLCache = {}

function load(mapfile)
  if mapCache[mapfile] then return mapCache[mapfile] end
  local mapdir = mapfile:gsub('[^/]*/?$', '')
  local xml = assert(FS.loadXML(mapfile))
  local map = assert(xmlChildren(xml).map, "no <map> element")
  map.width = tonumber(map.width)
  map.height = tonumber(map.height)
  map.tilewidth = tonumber(map.tilewidth)
  map.tileheight = tonumber(map.tileheight)
  map.maxtileheight = map.tileheight
  map.properties = {}
  map.layers = {}
  map.tiles = {}
  map.firstgids = {}
  local events = { onIntersect = true, onStand = true, onActivate = true }
  for i,v in ipairs(map) do
    if v.xml == 'properties' then
      loadProperties(map.properties, v, 'map')
    elseif v.xml == 'tileset' then
      local ts = v
      if ts.source then
        local firstgid = ts.firstgid
        ts = assert(FS.loadXML(mapdir..ts.source))
        ts = assert(xmlChildren(ts).tileset, "no <tileset> element")
        ts.firstgid = firstgid
      end
      map.firstgids[ts.name] = ts.firstgid
      local imageSource = assert((xmlChildren(ts).image or {}).source,
        "images embedded in the tileset aren't supported"):gsub('%-ed', '')
      ts.margin = tonumber(ts.margin or 0)
      ts.spacing = tonumber(ts.spacing or 0)
      ts.tilewidth = tonumber(ts.tilewidth or map.tilewidth)
      ts.tileheight = tonumber(ts.tileheight or map.tileheight)
      if ts.tileheight > map.maxtileheight then
        map.maxtileheight = ts.tileheight end
      for i,tile in ipairs(ts) do if tile.xml == 'tile' then
        local gid = ts.firstgid+tile.id
        map.tiles[gid] = { gid = gid }
        loadProperties(map.tiles[gid], xmlChildren(tile).properties,
          'tileset '..ts.name..' tile '..tile.id)
      end end
      local texture = Textures.load(mapdir..imageSource..'{^2}')
      local tw, th, i = ts.tilewidth, ts.tileheight, 0
      for y = ts.margin, texture.origh-ts.margin-1, th+ts.spacing do
        for x = ts.margin, texture.origw-ts.margin-1, tw+ts.spacing do
          map.tiles[ts.firstgid+i] = map.tiles[ts.firstgid+i] or {}
          map.tiles[ts.firstgid+i].w = ts.tilewidth
          map.tiles[ts.firstgid+i].h = ts.tileheight
          if not map.tiles[ts.firstgid+i].invisible then
            tilesetDLCache[imageSource] = tilesetDLCache[imageSource] or {}
            if not tilesetDLCache[imageSource][i] then
              tilesetDLCache[imageSource][i] = gl.GenLists(1)
              gl.NewList(tilesetDLCache[imageSource][i], GL.COMPILE)
                gl.BindTexture(GL.TEXTURE_2D, texture.name)
                gl.Begin(GL.QUADS)
                  -- upper left
                  gl.TexCoord(x/texture.w, (texture.origh-y)/texture.h)
                  gl.Vertex(0, th)
                  -- upper right
                  gl.TexCoord((x+tw)/texture.w, (texture.origh-y)/texture.h)
                  gl.Vertex(tw, th)
                  -- lower right
                  gl.TexCoord((x+tw)/texture.w, (texture.origh-y-th)/texture.h)
                  gl.Vertex(tw, 0)
                  -- lower left
                  gl.TexCoord(x/texture.w, (texture.origh-y-th)/texture.h)
                  gl.Vertex(0, 0)
                gl.End()
              gl.EndList()
            end
            map.tiles[ts.firstgid+i].display = tilesetDLCache[imageSource][i]
          end
          i = i + 1
        end
      end
    elseif v.xml == 'layer' then
      local layer = v
      layer.n = #map.layers+1
      map.layers[#map.layers+1] = layer
      map.layers[layer.name] = layer
      layer.x = tonumber(layer.x or 0)
      layer.y = tonumber(layer.y or 0)
      layer.width = tonumber(layer.width)
      layer.height = tonumber(layer.height)
      local children = xmlChildren(layer)
      layer.properties = loadProperties({}, children.properties,
        'layer '..layer.name)
      layer.data = nil
      if children.data.encoding == 'base64' then
        local encoded = children.data[1]:gsub('^%s+', ''):gsub('%s+$', '')
        local decoded = base64_decode(encoded)
        if children.data.compression == 'gzip' then
          local origLen = string.unpack(decoded:sub(-4), '<C')
          local origBuf = SDL.RWFromNewMem(origLen)
          origLen = zlib.uncompressZip(origBuf:baseptr(), origLen,
            decoded:sub(11, -9), #decoded-18)
          decoded = assert(origBuf:read(origLen))
        end
        layer.data = string.unpack(decoded,
          '<'..(layer.width*layer.height)..'C', true)
      else
        layer.data = {}
        for i,t in ipairs(children.data) do layer.data[i] = tonumber(t.gid) end
      end
      layer.fields = {}
      layer.absfields = {}
      local x, y = 0, 0
      for i,gid in ipairs(layer.data) do
        local mx, my = layer.x+x, layer.y+y
        layer.fields[y] = layer.fields[y] or {}
        layer.fields[y][x] = gid
        layer.absfields[my] = layer.absfields[my] or {}
        layer.absfields[my][mx] = gid
        x = x + 1; if x >= layer.width then x = 0; y = y + 1 end
      end
    elseif v.xml == 'objectgroup' then
      -- nothing for now
    end
  end
  map.script = assert(FS.loadText(mapfile:gsub('%.tmx$', '')..'.lua'))
  mapCache[mapfile] = map
  return map
end

