
local math, table = math, table
local SDL = require 'SDL'
local gl = require 'gl'; local GL = gl
local Fonts = require 'Fonts'
local Node = require 'Node'
local Textures = require 'Textures'
local Event = require 'Event'
local Mask = require 'Mask'

local assert, pairs, ipairs, notice = assert, pairs, ipairs, notice

module(...)

local MapView = _M
Node.Moving:subclass(MapView)

MapView.w = 0
MapView.h = 0

function MapView:displayStencil()
  gl.Begin(GL.QUADS)
    gl.Vertex(0, 0)
    gl.Vertex(0, self.h)
    gl.Vertex(self.w, self.h)
    gl.Vertex(self.w, 0)
  gl.End()
end

function MapView:displayContent()
  Mask.Stencil.push(self)

  local scroll_x, scroll_y = self.player.actor.x, -self.player.actor.y
  local map = self.player.map
  local health, maxHealth = self.player.health, self.player.maxHealth
  local message, warping = self.player.message, self.player.warping

  assert(map)
  local w2, h2 = math.floor(self.w/2), math.floor(self.h/2)
  assert(map.data.tilewidth%2 == 0, 'odd tilewidth')
  assert(map.data.tileheight%2 == 0, 'odd tileheight')
  local tw2, th2 = map.data.tilewidth/2, map.data.tileheight/2
  -- center the MapView on the center of tile (0,0)
  gl.PushMatrix()
  gl.Translate(w2-tw2, h2-th2, 0)
  gl.Translate(-scroll_x, scroll_y, 0)
  gl.Color(1, 1, 1)

  -- rules of drawing order:
  -- - a tile must be drawn after its top neighbors
  -- - an actor must be drawn after all the tiles it intersects
  -- - actors must be drawn in ascending orthogonal-Y coordinate order
  -- - an actor must be drawn before its bottom neighbors, they may occlude it

  -- visible area constraints
  -- ox, oy are orthogonal coords with one tile sized (2,2)
  -- ix, iy are isometric coords with one tile sized (1,1)
  local oy1 = math.floor((scroll_y-h2)/th2)
  local oy2 = math.ceil ((scroll_y+h2+map.data.maxtileheight)/th2)
  local ox1 = math.floor((scroll_x-w2)/tw2)
  local ox2 = math.ceil ((scroll_x+w2)/tw2)

  local world = self.parent.parent
  local cache = map.cache
  local time = world.time + self:getLoop().frameFraction

local did=0
  local alreadyDrawn = {}
  for i = 0, map.data.height-1 do alreadyDrawn[i] = 0 end
  local function displayField(lix, liy)
--    did=did+1
    -- display the field's top neighbors as well
    -- TODO: we may need to optimize these loops with o[xy][12] later
    for iy = 0, liy do
      for ix = alreadyDrawn[iy] or 0, lix do
--did=did+1
        local ox, oy = ix-iy, ix+iy
        if ox > ox2 or oy > oy2 then break end
        if not (ox < ox1 or oy < oy1) then
          for l = 1, #cache do
            local f = cache[l][iy] and cache[l][iy][ix]
            local color = f and f.color
            if f and f.frames then
              local begin = map.animTimes[l][iy][ix] or (map.initTime +
                (f.randbegin and (ix*601%127) * (iy*829%131) or 0))
              local frame = math.floor((time-begin)/f.delay)
              frame = f.noloop and math.min(frame, f.frames-1) or frame%f.frames
              f = map.data.tiles[f.gid + frame]
            end
            if f and f.display then
              gl.PushMatrix()
              if color then gl.Color(map.doorcolor[ix]) end
              gl.Translate(ox*tw2, -oy*th2, 0)
              gl.CallList(f.display)
              if color then gl.Color(1, 1, 1) end
              gl.PopMatrix()
            end
          end
        end
      end
      alreadyDrawn[iy] = math.max(alreadyDrawn[iy] or 0, lix + 1)
    end
  end

  -- display the actors in order, and drag in their 'dependencies'
  -- (sortedActors are sorted by the orthogonal-Y coordinate of the lower
  -- right corner, which is the actor's lowest point on the screen)
  for i,v in ipairs(map.sortedActors) do
    local fields = v:intersectedFields()
    displayField(fields[#fields][1], fields[#fields][2])
    --displayField(127, 127)
    v:display()
  end

  -- display all the other tiles
  displayField(map.data.width-1, map.data.height-1)
--notice(did)

  gl.PopMatrix()

  if world.frozen then warping = world.frozen end
  if warping then
    gl.BindTexture(GL.TEXTURE_2D, 0)
    gl.Color(0, 0, 0, 1-math.abs(warping)/100)
    gl.Begin(GL.QUADS)
      gl.Vertex(0, 0); gl.Vertex(self.w, 0)
      gl.Vertex(self.w, self.h); gl.Vertex(0, self.h)
    gl.End()
    gl.Color(1, 1, 1)
  end

  if message then
    local font = Fonts['fonts/ANTIQUEN.ttf:16']
    local bh = font.height * #message
    local bw = 0
    for i,v in ipairs(message) do bw = math.max(bw, font:prepare(v)) end
    bh = bh + 20; bw = bw + 20
    gl.BindTexture(GL.TEXTURE_2D, 0)
    gl.Color(0, 0, 0, 0.5)
    gl.Begin(GL.QUADS)
      gl.Vertex(w2-bw/2, h2-bh/2); gl.Vertex(w2-bw/2, h2+bh/2)
      gl.Vertex(w2+bw/2, h2+bh/2); gl.Vertex(w2+bw/2, h2-bh/2)
    gl.End()
    gl.Color(1, 1, 1)
    for i,v in ipairs(message) do
      gl.PushMatrix()
      gl.Translate(math.floor(w2-font:prepare(v)/2),
        math.floor(h2+font.height*(#message/2-i)), 0)
      font:print(v)
      gl.PopMatrix()
    end
  end

  if health then
    gl.PushMatrix()
    -- TODO don't use Textures.load in displayContent()
    local red = Textures.load('health.png')
    local black = Textures.load('healthblack.png')
    for i,v in ipairs({ red.name, black.name }) do
      gl.BindTexture(GL.TEXTURE_2D, v)
      if warping then gl.Color(1, 1, 1, math.abs(warping)/100) end
      gl.Begin(GL.QUADS)
        gl.TexCoord(0, 0); gl.Vertex(0, 0)
        gl.TexCoord(health, 0); gl.Vertex(health*16, 0)
        gl.TexCoord(health, 1); gl.Vertex(health*16, 16)
        gl.TexCoord(0, 1); gl.Vertex(0, 16)
      gl.End()
      gl.Translate(health*16, 0, 0)
      health = maxHealth - health
    end
    gl.PopMatrix()
  end

  Mask.Stencil.pop()
end

