
local assert, pairs, ipairs, notice = assert, pairs, ipairs, notice
local math = math
local Node = require 'Node'
local Map = require 'yarivan.Map'
local Actor = require 'yarivan.Actor'
local Projectile = require 'yarivan.Projectile'
local Event = require 'Event'

module(...)

local Player = _M
Node:subclass(Player)

Player.id = nil
Player.actor = nil

Player.key_left = nil
Player.key_right = nil
Player.key_up = nil
Player.key_down = nil
Player.key_activate = nil
Player.key_shoot = nil

Player.maxHealth = 5
Player.maxSpeed = 6
Player.acceleration = 0.8

Player.health = 5
Player.direction = 2
Player.speed = 0

function Player:setLocation(mapfile, x, y)
  local world = self.parent.parent
  local initializing = not self.map
  if not initializing then world:saveGame(10, '(Autosave)') end
  if not self.map or self.map.mapfile ~= mapfile then
    -- TODO for animated mapview transition, do this: 1. delete the actor,
    -- 2. fade out the mapview, 3. call setLocation, 4. fade in the mapview.
    -- (consider creating a copy of the actor when deleting it.)
    local old_actor = self.actor or {}
    self.map = world:getMap(mapfile) or
      Map:new{ mapfile = mapfile, parent = self }
    self.actor = PlayerActor:new{ parent = self.map, player = self,
      fade = old_actor.fade, alpha = old_actor.alpha }
    self.actor:loadTextures(self.actorTextures)
    self.map[#self.map+1] = self.actor
  end
  self.actor:setPosition(x, y)
  self.map:updateActorPositions()
  self.speed = 0
  if initializing then self.warping = 0 end
end

function Player:warp(mapfile, x, y)
  self.warpTo = { mapfile, x, y }
  self.warping = (self.warping and self.warping > 0) and self.warping or 100
  self.actor.ghost = true
  self.actor.fade = -0.2
  self.speed = 0
end

function Player:update()
  if self.immune then self.immune = self.immune - 1 end
  if self.immune == 0 then self.immune = nil end
  if self.reload then self.reload = self.reload - 1 end
  if self.reload == 0 then self.reload = nil end
  if self.dead then
    self.dead = self.dead - 1
    self:setMessage(math.floor(self.dead/25))
  end
  if self.dead == 0 then
    self.dead = nil; self.actor.ghost = nil; self.actor.fade = 0.1
    self.health = self.maxHealth; self:setMessage(nil)
  end
  if self.warping then self.warping = self.warping - 20 end
  if self.warping == 0 then
    self:setLocation(self.warpTo[1], self.warpTo[2], self.warpTo[3])
    self.warpTo = nil
    if not self.dead then self.actor.ghost = nil; self.actor.fade = 0.2 end
  end
  if self.warping == -100 then self.warping = nil end
  -- TODO we might want to move inertia to Actor
  -- TODO we might want to support actual vector speed and inertia
  -- (but that would require actual 45 degree rotation and sliding, so... no)
  self.speed = math.max(0, self.speed - self.acceleration)
  self.actor:move(self.direction, self.speed)
  -- check onStand
  local ourcenterx = self.actor.ix+self.actor.size/2
  local ourcentery = self.actor.iy+self.actor.size/2
  local fieldcenterx = math.floor(self.actor.ix)+0.5
  local fieldcentery = math.floor(self.actor.iy)+0.5
  local vx, vy = ourcenterx - fieldcenterx, ourcentery - fieldcentery
  if math.abs(vx) + math.abs(vy) < 0.3 then
    local fx, fy = math.floor(self.actor.ix), math.floor(self.actor.iy)
    if self.standx ~= fx or self.standy ~= fy then
      self.standx, self.standy = fx, fy
      self.map:callHook(fx, fy, 'onStand', self)
    end
  else
    self.standx, self.standy = nil
  end
end

function Player:processInput()
  -- call this *before* update()
  if self.dead then return end
  local dir = 22
  if Event.key[self.key_up]    then dir = dir - 10 end
  if Event.key[self.key_down]  then dir = dir + 10 end
  if Event.key[self.key_left]  then dir = dir - 1 end
  if Event.key[self.key_right] then dir = dir + 1 end
  if dir ~= 22 then
    self.direction = ({ [31] = 1, [32] = 2, [33] = 3, [23] = 4,
      [13] = 5, [12] = 6, [11] = 7, [21] = 8 })[dir]
    -- add 2a to counter for the natural deceleration
    self.speed = math.min(self.maxSpeed, self.speed + 2*self.acceleration)
  end
  if Event.key[self.key_activate] and self.didntActivate then
    local d = self.map.directionAngles[self.direction]
    self:activate(math.floor((self.actor.ix+self.actor.size/2)+d[1]/2),
      math.floor((self.actor.iy+self.actor.size/2)+d[2]/2))
  end
  self.didntActivate = not Event.key[self.key_activate]
  if Event.key[self.key_shoot] and not self.reload then
    local projectile = Projectile:new{ direction = self.direction,
      parent = self.map }
    projectile.ix = self.actor.ix+self.actor.size/2-projectile.size/2
    projectile.iy = self.actor.iy+self.actor.size/2-projectile.size/2
    projectile:loadTextures({ rel_y = 20, frames = 1, delay = 1,
      base = 'shot'..self.id..'/' })
    self.map[#self.map+1] = projectile
    self.reload = 7
  end
end

function Player:activate(x, y)
  self:setMessage(nil)
  self.map:callHook(x, y, 'onActivate', self)
end

function Player:directionBetween(a, b)
  if a <= b then
    return self.direction >= a and self.direction <= b
  else
    return self.direction >= a or self.direction <= b
  end
end

function Player:setMessage(message)
  if message then
    local t = {}
    for l in (message..'\n'):gmatch('.-\n') do t[#t+1] = l:sub(1,-2) end
    message = t
  end
  self.message = message
end

function Player:damage()
  if self.immune then return end
  self.health = self.health - 1
  self.immune = 5
  if self.health <= 0 then
    self.dead = 20*25
    self.actor.ghost = true
    self.actor.alpha = 1
    self.actor.fade = -0.1
    self.speed = 0
  else
    self.immune = 15
  end
end

function Player:throwBack(enemy)
  if self.actor.ghost then return end
  local dx, dy = self.actor.ix-enemy.ix, self.actor.iy-enemy.iy
  local dir = math.atan2(-dy, dx) * math.pi/4
  dir = (math.floor(dir + 0.5) + 11) % 8 + 1
  self.actor:throwBack(dir, 11, 4)
end

PlayerActor = Actor:subclass()

PlayerActor.size = 0.8

function PlayerActor:displayContent()
  Actor.displayContent(self)
  -- TODO display self.player.immune
end

function PlayerActor:movedInto(x, y)
  Actor.movedInto(self, x, y)
  --self.player:setMessage(nil)   -- stop displaying messages
  self.player.map:callHook(x, y, 'onIntersect', self.player)
end

