
require 'extlib'
require 'Node'
require 'Event'
require 'Textures'
require 'Selection'
require 'bit'
require 'gl'
require 'SDL'

local _G = _G
local bit, gl, glu, SDL = bit, gl, glu, SDL
local Event, Textures, Class, Selection =
  Event, Textures, Class, Selection

module(...)

local MainLoop = _M

Class(MainLoop)

MainLoop.frameFraction = 0

MainLoop.ticksPerSecond = 25
MainLoop.maxFrameskip = 5

function MainLoop:initialize()
  self.updateFps = _G.makeFpsSensor()
  self.root = self.root or _G.Node.Root:new{}
  self.focus = self.root
  self.root.mainloop = self
end

function MainLoop:__inspect(state)
  if not state.export then return _G.inspect.table(self, state) end
  return _G.inspect.table(self, state,
    _G.inspect.blacklist{ updateFps = true, nextUpdate = true })
end

function MainLoop:forceKeyRelease()
  Event.key_update()
  for sym, state in _G.pairs(Event.key) do if state ~= Event.RELEASED then
    ev = { sym = sym, mod = 0, unicode = 0, type = SDL.KEYUP, down = false }
    Event.key[sym] = Event.RELEASED; if self then self:key(ev) end
  end end
  Event.key_update()
end

function MainLoop:forceMouseRelease()
  if not self then return end
  for butid = 1, 5 do if Event.mouse[butid] then
    self:mouseButton({ type = SDL.MOUSEBUTTONUP, down = false, button = butid,
      x = Event.mouse.x, y = Event.mouse.y })
  end end
end

function MainLoop:quit()
  _G.error(false)  -- "already reported error", just end the program
end

function MainLoop:active(state, gain)
  Event.active[state] = gain
  --_G.notice("[active ".._G.tostring(state).." ".._G.tostring(gain).."]")
  
  if gain ~= false then return end   -- gains don't need to be handled here
  
  if bit.band(state, SDL.APPACTIVE) ~= 0 then
    -- the app also effectively loses mouse and keyboard focus
    self:active(SDL.APPINPUTFOCUS, false)
    self:active(SDL.APPMOUSEFOCUS, false)
    if self.pauseWhileMinimized then
      while true do   -- event loop while app is minimized
        local evn = SDL.PollEvent()
        if evn and evn.type == SDL.ACTIVEEVENT and evn.gain and
          bit.band(evn.state, SDL.APPACTIVE) ~= 0 then return end
        if evn and evn.type == SDL.QUIT then self:quit(); return end
        SDL.Delay(10)
      end
    end
  end
  
  if state == SDL.APPINPUTFOCUS then self:forceKeyRelease() end
  
  -- when the window loses mouse, drag'n'drop can still continue, so
  -- self:forceMouseRelease() isn't needed here. (tested under Win32 and X11)
end

function MainLoop:key(ev)
  ev.type = 'Key'; local received = Event.send(ev, self.focus)
  if ev.down and received then Event.key[ev.sym] = Event.OVERRIDDEN end
end

function MainLoop:mouseMotion(ev)
  if not self.forceHover then self.hover, self.hoverinfo =
    Selection.get(Event.mouse.x, _G.HEIGHT-Event.mouse.y, self) end
  if self.hover then self:setFocus(self.hover) end
  ev.type = 'MouseMotion'; Event.send(ev, self.focus)
end

function MainLoop:mouseButton(ev)
  if not self.forceHover then self.hover, self.hoverinfo =
    Selection.get(Event.mouse.x, _G.HEIGHT-Event.mouse.y, self) end
  if self.hover then self:setFocus(self.hover) end
  -- TODO remove MouseButtonAway, add 'background' nodes where applicable
  -- (and convert :eventMouseButton to check :getLoop().hover elsewhere)
  ev.type = self.hover and 'MouseButton' or 'MouseButtonAway'; Event.send(ev, self.focus)
end

function MainLoop:update()
  self.root:update()
end

function MainLoop:display()
  gl.Clear(gl.COLOR_BUFFER_BIT + gl.DEPTH_BUFFER_BIT)
  gl.LoadIdentity()
  self.root:display()
end

function MainLoop:setFocus(obj)
  if self.focus == obj then return end
  local oldfocus = self.focus; self.focus = obj
  -- when node:lostFocus() returns true, it overrides the focus change
  if oldfocus and oldfocus:lostFocus() then self.focus = oldfocus
  elseif self.focus then self.focus:gotFocus() end
end

function MainLoop:subLoop(klass, proto)
  proto = proto or {}
  proto.parentloop = self
  klass:new(proto):run()
  self.nextUpdate = SDL.GetTicks()
end

function MainLoop:replaceLoop(klass, proto)
  proto = proto or {}
  proto.parentloop = self.parentloop
  self.tail_fn = function () return klass:new(proto):run() end
  self.running = false
end

function MainLoop:run()
  local motion_ev = nil
  
  if self.parentloop then
    self.parentloop:forceKeyRelease(); self.parentloop:forceMouseRelease()
  else MainLoop.forceKeyRelease() end
  
  -- this game loop structure is modeled after the article at
  -- <http://dewitters.koonsolo.com/gameloop.html>
  
  self.nextUpdate = SDL.GetTicks()
  self.running = true
  while self.running do
    local loops = 0
    while SDL.GetTicks() > self.nextUpdate and loops < self.maxFrameskip do
      -- compress runs of SDL.MOUSEMOTION events - Selection.get() can be
      -- expensive and most of the time, only the last SDL.MOUSEMOTION matters
      local last_motion
      while true do   -- handle all SDL events in queue
        local ev = SDL.PollEvent()
        if ev and ev.type == SDL.MOUSEMOTION then   -- rewrite last_motion
          ev.buttons = nil   -- all button state changes should be
                             -- reported with SDL.MOUSEBUTTONUP/DOWN
          Event.mouse.x, Event.mouse.y = ev.x, ev.y
          last_motion = ev
        elseif last_motion then   -- this runs even when ev == nil
          self:mouseMotion(last_motion); last_motion = nil
        end
        if not ev then break end
        if ev.type == SDL.ACTIVEEVENT then
          for i,v in _G.ipairs({SDL.APPACTIVE, SDL.APPMOUSEFOCUS, SDL.APPINPUTFOCUS}) do
            if bit.band(ev.state, v) ~= 0 then self:active(v, ev.gain) end end
        elseif ev.type == SDL.QUIT then self:quit()
        elseif ev.type == SDL.KEYDOWN then
          ev.down = true; Event.key[ev.sym] = Event.JUST_PRESSED; self:key(ev)
        elseif ev.type == SDL.KEYUP then
          if Event.key[ev.sym] ~= Event.RELEASED then
            Event.key[ev.sym] = (Event.key[ev.sym] == Event.JUST_PRESSED
              and Event.JUST_RELEASED or Event.RELEASED)
            ev.down = false; self:key(ev) end
        elseif ev.type == SDL.MOUSEBUTTONDOWN then
          ev.x, ev.y = nil; ev.down = true
          Event.mouse[ev.button] = true; self:mouseButton(ev)
        elseif ev.type == SDL.MOUSEBUTTONUP then
          ev.x, ev.y = nil; ev.down = false
          if Event.mouse[ev.button] then
            Event.mouse[ev.button] = nil; self:mouseButton(ev) end
        end
      end
      
      self:update()
      Event.key_update()
      if not self.running then break end
      
      self.nextUpdate = self.nextUpdate + (1000 / self.ticksPerSecond)
      loops = loops + 1
    end
    if not self.running then break end
    
    -- frameFraction is a clampf representing what fraction of the current
    -- tick has passed. (update 'ticks' aren't the same as display 'frames'.)
    local skipTicks = 1000 / self.ticksPerSecond
    self.frameFraction = _G.math.min(1,
      (SDL.GetTicks() - self.nextUpdate + skipTicks) / skipTicks)
    
    SDL.PumpEvents()
    local _; _, Event.mouse.cur_x, Event.mouse.cur_y = SDL.GetMouseState()
    self:display(); SDL.GL_SwapBuffers()
    if self.forceDelay then SDL.Delay(self.forceDelay) end
    self.updateFps()
  end
  
  self:forceKeyRelease()
  self:forceMouseRelease()
  
  if self.tail_fn then return self:tail_fn() end
end

