
local _G = _G
local math, table, string = math, table, string
local ipairs, assert, setmetatable = ipairs, assert, setmetatable
local bit, SDL, gl, GL, glu = bit, SDL, gl, gl, glu

module(...)

local Selection = _M

Selection.active = false

---[==[   selection through color coding

-- To ease usage, this also hides or modifies GL functions that can interfere
-- with color coding (e.g. glShadeModel or glColor). However, this is a 'leaky
-- abstraction'. For example, glInterleavedArrays and glClearColor aren't
-- hidden. Take care when using this.

local Names, Stack, ColorMask

Selection.gl = nil

local rBits, gBits, bBits, rMask, gMask, bMask, rShift, gShift, bShift
local function setcolor(onlyMask)
  if not Stack[#Stack] then   -- when the stack's empty, don't draw anything
    Selection.gl.ColorMask(false, false, false, false); return
  else Selection.gl.ColorMask(_G.unpack(ColorMask, 1, 4)) end
  if onlyMask then return end
  -- from OpenGL FAQ 11.050
  if not rBits then
    rBits = gl.GetInteger(GL.RED_BITS,   1)
    gBits = gl.GetInteger(GL.GREEN_BITS, 1)
    bBits = gl.GetInteger(GL.BLUE_BITS,  1)
    -- glColor4ub won't work with more than 8 bits
    -- TODO: I'm not sure if the code will work correctly with *Bits>8
    -- (the following is supposed to make sure it does, but does it?)
    if rBits > 8 then rBits = 8 end
    if gBits > 8 then gBits = 8 end
    if bBits > 8 then bBits = 8 end
    rMask = bit.lshift(bit.lshift(1, rBits) - 1, gBits + bBits)
    gMask = bit.lshift(bit.lshift(1, gBits) - 1, bBits)
    bMask =            bit.lshift(1, bBits) - 1
    rShift = rBits + gBits + bBits - 8
    gShift =         gBits + bBits - 8
    bShift =                 bBits - 8
  end
  _G.assert(Stack[#Stack] < bit.lshift(1, rBits + gBits + bBits) - 1,
    "too many selectable objects")
  Selection.gl.Color4ub(
    bit.rshift(bit.band(Stack[#Stack], rMask), rShift),
    bit.rshift(bit.band(Stack[#Stack], gMask), gShift),
    bit.rshift(bit.band(Stack[#Stack], bMask), bShift), 255)
end
local function color2id(r, g, b, a)
  if not rBits then return nil end
  if r == 255 and g == 255 and b == 255 then return nil end
  return bit.bor(
    bit.band(bit.lshift(r, rShift), rMask),
    bit.band(bit.lshift(g, gShift), gMask),
    bit.band(bit.lshift(b, bShift), bMask))
end

function Selection.push(object, ...)
  if not Selection.active then return end
  local name = #Names + 1
  Names[name] = { object = object, ... }
  Stack[#Stack+1] = name; setcolor()
end

function Selection.pop()
  if not Selection.active then return end
  Stack[#Stack] = nil; setcolor()
end

local mybuf = SDL.RWFromNewMem(4)
function Selection.get(x, y, selected)
  gl.PushAttrib(GL.ALL_ATTRIB_BITS)
  -- from OpenGL FAQ 11.040
  local bad_modes = { [GL.BLEND] = true, [GL.DITHER] = true, [GL.FOG] = true,
    [GL.LIGHTING] = true, [GL.TEXTURE_1D] = true, [GL.TEXTURE_2D] = true }
  if GL.TEXTURE_3D then bad_modes[GL.TEXTURE_3D] = true end   -- TODO is 'if' needed?
  for k, v in _G.pairs(bad_modes) do gl.Disable(k) end
  gl.ShadeModel(GL.FLAT)
  gl.ClearColor(1, 1, 1, 1); gl.Color(1, 1, 1, 1)
  if not Selection.gl then   -- create copies of functions we'll hide
    local block = { Color = true, ColorPointer = true, ColorMask = true,
      ShadeModel = true, Enable = true }
    Selection.gl = {}
    for k, v in _G.pairs(gl) do
      if block[k] or k:match('^Color%d') then Selection.gl[k] = v end end
  end
  for k, v in _G.pairs(Selection.gl) do gl[k] = function () end end
  gl.Enable = function (v)
    return not bad_modes[v] and Selection.gl.Enable(v) or nil end
  gl.ColorMask = function (...) ColorMask = {...}; setcolor(true) end
  ColorMask = gl.GetBoolean(GL.COLOR_WRITEMASK, 4); Names = {}
  Stack, active = {}, true
  setcolor(); selected:display()
  Stack, active = nil, false
  for k, v in _G.pairs(Selection.gl) do gl[k] = v end   -- restore gl funcs
  gl.PopAttrib()
  
  gl.ReadPixels(x, y, 1, 1, GL.RGBA, GL.UNSIGNED_BYTE, mybuf:baseptr())
  local id = color2id(string.unpack(mybuf:baseptr(), mybuf:stopptr(), '4A'))
  local nameInfo = Names[id]; Names = nil
  return nameInfo and nameInfo.object, nameInfo
end

--[===[ ]==]

-- classic OpenGL selection with glSelectBuffer
-- This model can detect all objects that were selected, not just the one on
-- top, but it was abandoned because it can't handle tests (e.g. stencil test
-- or scissor test), and false positives were reported for invisible objects.

Selection.bufferSize = 64   -- TODO not enough?

local pickbuf
local Names

function push(object, ...)
  if not Selection.active then return end
  local name = #Names + 1
  Names[name] = { object = object, ... }
  gl.PushName(name)
end

function pop()
  gl.PopName()
end

function getHits(x, y, selected)
  local zero = string.pack('I', 0)
  if not pickbuf then
    pickbuf = SDL.RWFromNewMem(bufferSize*#zero) end
  assert(pickbuf:seek('set'))   -- clear the buffer
  assert(pickbuf:write(zero:rep(bufferSize)))
  
  gl.MatrixMode(GL.PROJECTION)
  local projection = gl.GetDouble(GL.PROJECTION_MATRIX)
  gl.LoadIdentity()
  glu.PickMatrix(x, y, 1, 1, gl.GetInteger(GL.VIEWPORT))
  gl.MultMatrix(projection)
  gl.MatrixMode(GL.MODELVIEW)
  
  gl.SelectBuffer(bufferSize, pickbuf:baseptr()); gl.RenderMode(GL.SELECT)
  Names = Names or {}; active = true; selected:display(); active = false
  local len = gl.RenderMode(gl.RENDER)
  local content = string.unpack(pickbuf:baseptr(), pickbuf:stopptr(),
    bufferSize..'I', true)
  
  local hits = {}
  local pos = 1
  if len < 0 then len = math.huge end
  for i = 1, len do
    if not content[pos+2] then break end  -- incomplete hit
    if not content[pos+2+content[pos]] then break end  -- incomplete hit
    local hit = { minz = content[pos+1], maxz = content[pos+2] }
    for j = 1, content[pos] do hit[j] = content[pos+2+j] end
    table.insert(hits, hit)
    pos = pos + 3 + content[pos]
  end
  
  gl.MatrixMode(GL.PROJECTION); gl.LoadMatrix(projection)
  gl.MatrixMode(GL.MODELVIEW)
  
  return hits
end

function get(x, y, selected)
  Names = {}
  local hits = getHits(x, y, selected)
  table.sort(hits, function (a,b) return a.minz < b.minz end)
  -- (TODO: the above should sort hits so that hits[1] is nearest. check it.)
  local nameInfo
  for i, hit in ipairs(hits) do
    -- ignore empty (unnamed) hits
    if hit[1] then nameInfo = Names[hit[#hit]]; break end
  end
  Names = nil
  return nameInfo and nameInfo.object, nameInfo
end


--]===]
