
local assert, pairs, ipairs, notice = assert, pairs, ipairs, notice
local gl = require 'gl'; local GL = gl
local SDL = require 'SDL'
local FS = FS
local math = math
local tonumber = tonumber

module(...)

-- returns src resized to have dimensions of values 2^n.
local function nearestPowerOf2(n)
  local p = 2; while p < n do p = p*2 end; return p
end
function resizeAsTexture(src)
  local w, h = src:w(), src:h()
  local newsize = nearestPowerOf2(math.max(w, h))
  local dest = SDL.CreateRGBSurface(newsize, newsize)
  dest:FillRect(0, 0, 0); src:BlitSurface(nil, dest, { 0, newsize - h })
  return dest
end

function generateAndBindName()
  local name = gl.GenTextures(1)
  gl.BindTexture(GL.TEXTURE_2D, name)
  gl.TexParameter(GL.TEXTURE_2D, GL.TEXTURE_MIN_FILTER, GL.LINEAR)
  gl.TexParameter(GL.TEXTURE_2D, GL.TEXTURE_MAG_FILTER, GL.LINEAR)
  return name
end

imageCache = {}

function load(texid)
  if not texid then
    -- In early versions of Scion, it was common for missing textures to crash
    -- the whole program. This is a failsafe that binds texture 0. It won't
    -- display correctly, but at least it won't fail ungracefully.
    notice("WARNING: got nil texid! binding 0.")
    return { name=0, w=1, h=1, origw=1, origh=1, wratio=1, hratio=1 }
  end
  if _M[texid] then return _M[texid] end

  local flags = texid:match('{[^}]*}$') or ''
  local fname = texid:sub(1, -1-#flags)
  flags = flags:match('^{(.*)}$') or ''
  local mipmap = flags:match('!!$'); flags = flags:gsub('!!$', '')
  if not imageCache[fname] then
    imageCache[fname] = assert(FS.loadImage(fname))
    imageCache[fname]:SetAlpha(0, 255)
  end
  local image = imageCache[fname]

  local texture = {}
  texture.origw, texture.origh = image:w(), image:h()
  local resize = nil
  if flags == '^2' then
    image = resizeAsTexture(image)
  elseif flags:match('^%d+,%d+,%d+,%d+$') then
    local oldimage = image
    local x, y, w, h = flags:match('^(%d+),(%d+),(%d+),(%d+)$')
    image = SDL.CreateRGBSurface(w, h)
    oldimage:BlitSurface({ x, y, w, h }, image, { 0, 0 })
  elseif flags:match('^%*[0-9.]+$') then
    -- *(number) makes Textures.load pretend that the texture has other size
    -- than what it really has. this only changes texture.{[wh],orig[wh]}, but
    -- that's enough for computing glTexCoords.
    -- TODO: find out what this is used for. consider removing it.
    resize = assert(tonumber(flags:sub(2)))
    texture.origw = math.floor(image:w() * resize)
    texture.origh = math.floor(image:h() * resize)
  else
    assert(flags == '', "unknown texture flags")
  end
  texture.w, texture.h = image:w(), image:h()
  if resize then texture.w, texture.h = texture.origw, texture.origh end
  texture.wratio = texture.origw / texture.w
  texture.hratio = texture.origh / texture.h

  texture.name = generateAndBindName()
  image:glTexImage2D(not not mipmap)

  _M[texid] = texture; _M[texture.name] = texture
  return texture
end

