
require 'Node'

local _G = _G
local ipairs = ipairs
local Class, Mixin = Class, Mixin

module(...)


local Widget = _M
_G.Node:instantiate(Widget)

function Widget:set_mixin(...)
  local parents = { ... }
  parents[#parents+1] = _M
  Mixin(self, unpack(parents))
end

function Widget:initialize()
  _G.Node.initialize(self)
  self.geometry = { 0, 0, 0, 0 }
  self.layout = nil
end

function Widget:updateGeometry()
  if self.layout then return self.layout:updateGeometry() end
end


Layout = Class()

function Layout:initialize()
  _G.assert(self.widget, "Widget.Layout:initialize() requires self.widget")
  self.rules = { }
end


HBoxLayout = Layout:instantiate()

local function _computeGeometry(self, min_sizes, rules, total_avail)
  local wi = self.widget
  local avail, factors_sum, one_part, everything_ok
  local sizes = {}
  while not everything_ok do
    local avail = total_avail
    factors_sum = 0
    for i,v in ipairs(wi.children) do
      if rules[i] == nil then sizes[i] = min_sizes[i] end
      avail = avail - (rules[i] ~= nil and 0 or min_sizes[i])
      factors_sum = factor_sum + (rules[i] or 0)
    end
    if avail < 0 then
      -- this happens when total_avail is not enough and it's below
      -- the minimum size of wi. content overflow will probably result
      -- (meaning, the widget's contents will be outside its bounding box)
      return min_sizes  -- "set" all sizes to min_sizes
    end
    one_part = (factors_sum == 0 and 0 or avail / factors_sum)
    everything_ok = true
    for i,v in ipairs(wi.children) do if rules[i] ~= nil then
      sizes[i] = one_part * rules[i]
      if sizes[i] < min_sizes[i] then
        everything_ok = false
        rules[i] = nil   -- rule==nil means use min_size
      end
    end end
    -- TODO: can this become an endless cycle?
  end
  return sizes
end

function HBoxLayout:minimumWidth()
  local wi = self.widget
  local inw = 0
  for i,v in ipairs(wi.children) do
    inw = inw + wi.spacing + v:minimumWidth()
  end
  return wi.padding[1] + (inw - wi.spacing) + wi.padding[3]
end

function HBoxLayout:minimumHeight(w)
  local wi = self.widget
  local inh = 0
  -- minimumHeight() requires the widget height as an argument, so
  -- this has to compute widths of all child widgets
  local min_sizes, min_rules = {}, {}
  for i,v in ipairs(wi.children) do min_sizes[i] = v:minimumWidth() end
  local total_avail = w - wi.padding[1] - wi.padding[3]
  local sizes = _computeGeometry(self, min_sizes, min_rules, total_avail)
  for i,v in ipairs(wi.children) do
    local size = v:minimumHeight(sizes[i])
    if inh < size then inh = size end
  end
  return wi.padding[2] + inh + wi.padding[4]
end

function HBoxLayout:updateGeometry()
  local wi = self.widget
  local inw = wi.geometry[3] - wi.padding[1] - wi.padding[3]
  local inh = wi.geometry[4] - wi.padding[2] - wi.padding[4]
  local total_avail = inw - wi.spacing * (#wi.children > 1 and #wi.children-1 or 0)
  local min_sizes = {}
  for i,v in ipairs(wi.children) do min_sizes[i] = v:minimumWidth() end
  local sizes = _computeGeometry(self, min_sizes, _G.table.icopy(self.rules), total_avail)
  local pos = wi.padding[1]
  for i,v in ipairs(wi.children) do
    v:setGeometry(_G.math.floor(pos), wi.padding[2], _G.math.floor(sizes[i]), inh)
    pos = pos + sizes[i] + wi.spacing
  end
end


VBoxLayout = Layout:instantiate()

function VBoxLayout:minimumWidth()
  local wi = self.widget
  local inw = 0
  for i,v in ipairs(wi.children) do
    local size = v:minimumWidth()
    if inw < size then inw = size end
  end
  return wi.padding[1] + inw + wi.padding[3]
end

function VBoxLayout:minimumHeight(w)
  local wi = self.widget
  local inw = w - wi.padding[1] - wi.padding[3]
  local inh = 0
  for i,v in ipairs(wi.children) do
    inh = inh + wi.spacing + v:minimumHeight(inw)
  end
  return wi.padding[2] + (inh - wi.spacing) + wi.padding[4]
end

function VBoxLayout:updateGeometry()
  local wi = self.widget
  local inw = wi.geometry[3] - wi.padding[1] - wi.padding[3]
  local inh = wi.geometry[4] - wi.padding[2] - wi.padding[4]
  local total_avail = inh - wi.spacing * (#wi.children > 1 and #wi.children-1 or 0)
  local min_sizes = {}
  for i,v in ipairs(wi.children) do min_sizes[i] = v:minimumHeight(inw) end
  local sizes = _computeGeometry(self, min_sizes, _G.table.icopy(self.rules), total_avail)
  local pos = wi.padding[2]
  for i,v in ipairs(wi.children) do
    v:setGeometry(wi.padding[1], _G.math.floor(pos), inw, _G.math.floor(sizes[i]))
    pos = pos + sizes[i] + wi.spacing
  end
end



