
-- EXTENDED LIBRARY

-- function notice()

function notice(msg, ...)
  local ar = debug.getinfo(2, 'Sl')
  local ln = ar.currentline and ar.currentline > 0 and
    ar.short_src..":"..ar.currentline..": " or ""
  io.write("notice: "..ln..tostring(msg))
  local len = select('#', ...)
  for i = 1, len do
    io.write(" "..inspect.object(select(i, ...), inspect.indented)) end
  io.write"\n"
end

-- deep table copy

function table.dcopy(src)
  local dest = {}
  for k,v in pairs(src) do dest[k] = (type(v) == 'table' and table.dcopy(v) or v) end
  return dest
end

-- FPS counter

function makeFpsSensor(auto_display)
  local last_update, frames = 0, 0
  return function ()
    -- this should be called every frame
    frames = frames + 1
    local now = SDL.GetTicks()
    local sec_passed = (now - last_update) / 1000
    if math.floor(now / 1000) <= math.floor(last_update / 1000) then
      return end   -- the next second hasn't begun yet
    local msg = ("%2ds %.3ffps"):format(now/1000 % 100, frames/sec_passed)
    last_update = now; frames = 0
    if auto_display == false then return msg end
    if _WIN32 then notice(msg); return msg end
    -- when not in Win32, it's assumed that the output goes to a terminal.
    -- the following escape sequences were tested on an xterm.
    local amsg = "   "..msg; io.stderr:write("\0277\027[A\027[512G"..
      amsg:gsub('.', '\b')..amsg.."\0278")
    return msg
  end
end

--------------------------------

-- additional string functions

function string.chomp(s, sep)
  return s:gsub((sep or '\r?\n\r?')..'$', '')
end

do
  local function make_quote_subst(t, ch)
    t[ch] = string.format('\\%03d', ch:byte()); return t[ch]
  end
  local quote_subst = { ['\b'] = '\\b', ['\n'] = '\\n', ['\r'] = '\\r',
    ['\t'] = '\\t', [']'] = ']', [ [[\]] ] = [[\\]], ['"'] = [[\"]] }
  setmetatable(quote_subst, { __index = make_quote_subst })
  
  function string.quote(s)
    return '"' .. s:gsub("[^ !#-[^-~]", quote_subst) .. '"'
  end
end

if getmetatable("") then getmetatable("").__inspect = string.quote end

function string.ufix(s)
  return (' '..s..' '):gsub('[\127\254\255]', '?')
    :gsub('[\192-\253]+([^\128-\191])', '?%1')
    :gsub('([^\128-\253])[\128-\191]+', '%1?')
    :gsub('[\192-\253][\128-\191]+', function (q) local b = string.byte(q)
      return (#q ~= (b >= 252 and 6 or b >= 248 and 5 or b >= 240 and 4 or
	b >= 224 and 3 or 2) and '?') end):sub(2, -2)
end

function string.ulen(s)
  return #s:ufix():gsub('[\128-\191]', '')
end

function string.usub(s, i, j)
  s = s:ufix(); local l = s:ulen()
  j = j or -1; if i < 0 then i = l+1+i end; if j < 0 then j = l+1+j end
  local n, vv, st = 0, 0, nil
  if i > l or j < 0 or i > j then return '' end
  for p in s:gmatch('[%0\1-\127\192-\255][\128-\191]*') do
    n = n + 1
    if n == i then st = vv + 1 end
    vv = vv + #p
    if n == j then return s:sub(st, vv) end
  end
  return s:sub(st, vv)
end

function string.uchar(...)
  local st, len, pr = {}, nil, nil
  for _,n in ipairs({...}) do
    len, pr = nil, nil
    if n > 0x04000000 then len, pr = 6, 0xFC
    elseif n > 0x00200000 then len, pr = 5, 0xF8
    elseif n > 0x00010000 then len, pr = 4, 0xF0
    elseif n > 0x00000800 then len, pr = 3, 0xE0
    elseif n > 0x00000080 then len, pr = 2, 0xC0
    else len, pr = 1, 0 end
    for i = len-1, 0, -1 do
      local m = bit.rshift(n, 6*i)
      st[#st+1] = i == len-1 and (pr + m) or (0x80 + bit.band(0x3F, m))
    end
  end
  return string.char(unpack(st))
end

--------------------------------

-- base64 decoder

do
  local decode_t = {}
  for i = 1, 26 do decode_t[i + 64] = i - 1 end
  for i = 1, 26 do decode_t[i + 96] = i - 1 + 26 end
  for i = 0, 9 do decode_t[i + 48] = i + 26 + 26 end
  decode_t[43] = 26 + 26 + 10
  decode_t[47] = 26 + 26 + 10 + 1
  decode_t[61] = 0
  
  function base64_decode(s)
    _G.assert(#s%4 == 0, "unsupported base64 length")
    local out = {}
    for i = 0, #s-1, 4 do
      local m = bit.bor(
        bit.lshift(decode_t[string.byte(s, i+1, i+1)], 18),
        bit.lshift(decode_t[string.byte(s, i+2, i+2)], 12),
        bit.lshift(decode_t[string.byte(s, i+3, i+3)], 6),
        bit.lshift(decode_t[string.byte(s, i+4, i+4)], 0))
      table.insert(out, string.char(
        bit.band(bit.rshift(m, 16), 0xFF),
        bit.band(bit.rshift(m, 8), 0xFF),
        bit.band(m, 0xFF)))
    end
    return table.concat(out)
  end
end

--------------------------------

-- the 'inspect' functions

-- NOTE: Don't call inspect.object(self) from inside __inspect! It will cause
-- a stack overflow.

inspect = {}

function inspect.userdata(o, state)
  assert(not state.export, "userdata can't be exported")
  local typename = nil
  if debug and debug.getregistry then
    local metatable = getmetatable(o)
    for n, v in pairs(debug.getregistry()) do
      if v == metatable then typename = n end
    end
  end
  if not typename then return "<"..tostring(o)..">" end
  return "<"..typename.." - "..tostring(o)..">"
end

inspect['function'] = function (o, state)
  assert(not state.export, "functions can't be exported")
  if not debug or not debug.getinfo then return "<"..tostring(o)..">" end
  local typename = nil
  local ar = debug.getinfo(o, "Sn")
  local str = "<"..ar.what.." "..tostring(o)
  if ar.namewhat and ar.namewhat ~= "" then
    return str .. " ('"..ar.namewhat.."')>"
  elseif ar.short_src and ar.short_src ~= "" and ar.what ~= "C" then
    return str .. " ("..ar.short_src..":"..ar.linedefined..")>"
  else
    return str .. ">"
  end
end

function inspect.thread(o, state)
  assert(not state.export, "threads can't be exported")
  return "<"..tostring(o)..">"
end

function inspect.string(o)
  -- string.__inspect is normally string.quote, but this is a fallback
  return string.format("%q", o)
end

function inspect.iterate(t)
  local ipairs_phase = true
  return function (t, k)
    if ipairs_phase then
      local i = k and k+1 or 0
      if rawget(t, i) then return i end
      ipairs_phase = false
    end
    return next(t, k)
  end, t
end

function inspect.blacklist(blacklist)
  return function (t)
    local orig_iter = inspect.iterate(t)
    return function (t, k)
      repeat k = orig_iter(t, k) until not blacklist[k]
      return k
    end, t
  end
end

function inspect.whitelist(whitelist)
  return function (t)
    return function (t, k)
      repeat k = next(whitelist, k) until not k or
        (whitelist[k] and rawget(t, k) ~= nil)
      return k
    end, t
  end
end

--[[  About indentation:
    state.indent is what gets written before every table key-value pair.
      When indentation is off, it's one space. When indentation is on, it's
      a newline and some spaces (depending on current indentation level).
    state.indent_inc is what gets appended to indent in every recursion level
      of _inspect.table. When indentation is off, it's nothing. When
      indentation is on, it's the indentation character (e.g. "\t").
  ]]

function inspect.table(o, state, iterator)
  iterator = iterator or state.iterator or inspect.iterate
  local result = '{'
  
  state.inspecting = state.inspecting or {}
  if state.inspecting[o] then
    assert(not state.export, "recursive tables can't be exported")   -- TODO
    return '<...>' end   -- prevent infinite loop
  state.inspecting[o] = true
  if getmetatable(o) and not state.export then
    result = '{ --[[ has a metatable ]]' end
  
  local keys_done, empty, last_int = {}, true, nil
  local function try_key(k)
    if keys_done[k] then return end; keys_done[k] = true
    local show_key; empty = false   -- from Lua manual:
    if (k == 1 and not last_int) or (last_int and last_int + 1 == k) then
      last_int = k; show_key = ''   -- field ::= exp
    elseif type(k) == 'string' and k:match('^[%a_][%w_]*$') then
      show_key = k..' = '           -- field ::= Name '=' exp
    else                            -- field ::= '[' exp ']' '=' exp
      show_key = '['..inspect.object(k, state)..'] = '
    end
    result = result .. state.indent .. show_key ..
      inspect.object(rawget(o, k), state) .. ','
  end
  
  state.indent = state.indent or " "; local old_indent = state.indent
  state.indent = state.indent .. (state.indent_inc or '')   -- add indentation
  for k in iterator(o) do try_key(k) end
  state.indent = old_indent   -- remove indentation
  
  if not state.concise then state.inspecting[o] = false end
  if empty then return result .. '}' end
  -- remove the last comma and add a closing bracket
  return result:gsub(',$', '') .. state.indent .. '}'
end

function inspect.object(o, state)
  return (rawget(getmetatable(o) or {}, "__inspect") or
    inspect[type(o)] or tostring)(o, state or {})
end

-- inspection modes/states
-- inspect.indented creates a new table every time (if the state is in
-- a global variable, inspect.object(_G) may recurse infinitely)
setmetatable(inspect, { __index = function (t, k)
  if k == 'indented' then return { indent = "\n", indent_inc = "  " } end
  if k == 'export' then return { indent = "\n", indent_inc = "  ",
    export = true } end
  return nil
end })

