
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <SDL_opengl.h>

#include <ft2build.h>
#include FT_FREETYPE_H

#include "ftgl.h"

struct _FTGL_Glyph {
  long tex_x, tex_y;   // glyph coordinates on the texture (in pixels)
  double l, r, u, d;  // glyph coordinates (in texture coordinates) - left, right, up, down
  long w, h;
  long offleft, offtop, advance; // bitmap offset (left, top) and glyph advance
  long character;     // UCS character code of the glyph
  FT_UInt ft_index;      // Freetype internal glyph index (this isn't the same thing as gid)
};

static int increase_texture_size(FTGL_Font* font) {
  int y;
  int new_tex_size = font->texture_size * 2;
  int new_pitch, old_pitch;
  unsigned char* new_tex_data = (unsigned char *)malloc(2 * new_tex_size * new_tex_size);
  if(!new_tex_data) {
    free((void *)font->texture_data); font->texture_data = NULL;
    return 0;
  }
  memset(new_tex_data, 0, 2 * new_tex_size * new_tex_size);
  new_pitch = new_tex_size * 2; old_pitch = font->texture_size * 2;
  for(y = 0; y < font->texture_size; y++)
    memcpy(&new_tex_data[new_pitch*y], &font->texture_data[old_pitch*y], old_pitch);  // copy one line from the old texture
  free((void *)font->texture_data);
  font->texture_size = new_tex_size;
  font->texture_data = new_tex_data;
  return 1;
}

static int render_glyph(FTGL_Font* font, long ch) {
  int gx, gy, w, tex_pos;
  long gid;
  FT_UInt ft_index;
  FT_GlyphSlot gslot;     // glyph slot, provided by Freetype
  struct _FTGL_Glyph* glyph;  // glyph data inside struct font_t
  
  ft_index = FT_Get_Char_Index(font->face, ch);
  
  if(FT_Load_Glyph(font->face, ft_index, FT_LOAD_RENDER) != 0)
    return 1;   // ignore errors, only don't load the glyph
  
  if(font->num_glyphs == font->num_allocated_glyphs) {   /* allocate more glyphs */
    font->num_allocated_glyphs += 8;
    font->glyph = (struct _FTGL_Glyph *)realloc(font->glyph, font->num_allocated_glyphs * sizeof(struct _FTGL_Glyph));
    if(!font->glyph) return 0;
    memset(font->glyph + font->num_glyphs, 0, (font->num_allocated_glyphs - font->num_glyphs) * sizeof(struct _FTGL_Glyph));
  }
  
  gid = font->num_glyphs;
  font->num_glyphs++;
  
  gslot = font->face->glyph;
  w = gslot->bitmap.width + 2;
  if(font->pen_x + w > font->texture_size) {
    font->pen_x = 0; font->pen_y += font->height;
  }
  if(font->pen_y + font->height > font->texture_size) {
    if(increase_texture_size(font) != 1) return 0;   // need bigger texture
  }
  glyph = &font->glyph[gid];
  glyph->w = gslot->bitmap.width; glyph->h = gslot->bitmap.rows;
  glyph->tex_x = font->pen_x; glyph->tex_y = font->pen_y;
  glyph->offleft = gslot->bitmap_left; glyph->offtop = gslot->bitmap_top - glyph->h;
  glyph->advance = gslot->advance.x >> 6; glyph->character = ch;
  glyph->ft_index = ft_index;
  for(gy = 0; gy < glyph->h; gy++) {
    for(gx = 0; gx < glyph->w; gx++) {
      tex_pos = 2 * (font->pen_x+gx + font->texture_size*(font->pen_y+gy));  // bpp*(x+w*y)
      // texture is GL_LUMINANCE_ALPHA. luminance is always 1
      font->texture_data[tex_pos] = 255;
      font->texture_data[tex_pos + 1] = gslot->bitmap.buffer[gx + gy*gslot->bitmap.pitch];
    }
  }
  font->pen_x += w;
  return 1;
}

static void save_texture(FTGL_Font* font) {
  int i;
  if(font->texture)
    glBindTexture(GL_TEXTURE_2D, font->texture);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, font->texture_size, font->texture_size, 0, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, font->texture_data);
  /* re-compute texture coordinates of glyphs */
  for(i = 0; i < font->num_glyphs; i++) {
    font->glyph[i].l = 1.0 / font->texture_size * (font->glyph[i].tex_x);
    font->glyph[i].r = 1.0 / font->texture_size * (font->glyph[i].tex_x + font->glyph[i].w);
    font->glyph[i].u = 1.0 / font->texture_size * (font->glyph[i].tex_y);
    font->glyph[i].d = 1.0 / font->texture_size * (font->glyph[i].tex_y + font->glyph[i].h);
  }
}

static long get_kerning(FTGL_Font *font, int prev_gid, int gid) {
  FT_Vector delta;
  if(!(font->use_kerning && prev_gid && gid))
    return 0;
  FT_Get_Kerning(font->face, font->glyph[prev_gid].ft_index,
      font->glyph[gid].ft_index, FT_KERNING_DEFAULT, &delta);
  return (delta.x >> 6);
}

static void draw_underline(FTGL_Font *font, double x1, double x2) {
  int y = FT_MulFix(font->face->underline_position, font->face->size->metrics.y_scale) >> 6;
  glPushAttrib(GL_ENABLE_BIT);
  glDisable(GL_TEXTURE_2D);
  glBegin(GL_LINES); glVertex2d(x1, y); glVertex2d(x2, y); glEnd();
  glPopAttrib();
}

static long utf8_to_ucs(const char* str) {
  unsigned char ch = *((const unsigned char *)str);
  if(ch < 0xC0)
    return ch;
  if(ch < 0xE0) {
    if(!str[1]) return ch;
    return (((long)(str[0] & 0x1F)) << 6)  |  ((long)(str[1] & 0x3F));
  }
  if(ch < 0xF0) {
    if(!str[1] || !str[2]) return ch;
    return (((long)(str[0] & 0x1F)) << 12) | (((long)(str[1] & 0x3F)) << 6) |
            ((long)(str[2] & 0x3F));
  }
  if(ch < 0xFA) {
    if(!str[1] || !str[2] || !str[3]) return ch;
    return (((long)(str[0] & 0x07)) << 18) | (((long)(str[1] & 0x3F)) << 12) |
           (((long)(str[2] & 0x3F)) << 6)  |  ((long)(str[3] & 0x3F));
  }
  if(ch < 0xFC) {
    if(!str[1] || !str[2] || !str[3] || !str[4]) return ch;
    return (((long)(str[0] & 0x07)) << 24) | (((long)(str[1] & 0x3F)) << 18) |
           (((long)(str[2] & 0x3F)) << 12) | (((long)(str[3] & 0x3F)) << 6)  |
            ((long)(str[4] & 0x3F));
  }
  if(ch < 0xFE) {
    if(!str[1] || !str[2] || !str[3] || !str[4] || !str[5]) return ch;
    return (((long)(str[0] & 0x07)) << 30) | (((long)(str[1] & 0x3F)) << 24) |
           (((long)(str[2] & 0x3F)) << 18) | (((long)(str[3] & 0x3F)) << 12) |
           (((long)(str[4] & 0x3F)) << 6)  |  ((long)(str[5] & 0x3F));
  }
  return ch;
}

FTGL_Font* FTGL_Create(FT_Face face, int initial_texture_size, unsigned int texture_id, int use_kerning) {
  int ch;
  FTGL_Font* font = (FTGL_Font *)malloc(sizeof(FTGL_Font));
  if(!font) return NULL;
  memset(font, 0, sizeof(*font));
  font->face = face; font->texture = texture_id;
  font->texture_size = initial_texture_size;
  font->height = (face->size->metrics.height + 63) >> 6;
  font->ascent = (FT_MulFix(face->bbox.yMax, face->size->metrics.y_scale) + 63) >> 6;
  font->descent = FT_MulFix(face->bbox.yMin, face->size->metrics.y_scale) >> 6;
  font->use_kerning = use_kerning && FT_HAS_KERNING(face);
  font->num_allocated_glyphs = 96;
  font->glyph = (struct _FTGL_Glyph *)malloc(font->num_allocated_glyphs * sizeof(struct _FTGL_Glyph));
  if(!font->glyph) { FTGL_Destroy(font); return NULL; }
  memset(font->glyph, 0, font->num_allocated_glyphs * sizeof(struct _FTGL_Glyph));
  font->texture_data = (unsigned char *)malloc(2 * initial_texture_size * initial_texture_size);
  if(!font->texture_data) { FTGL_Destroy(font); return NULL; }
  memset(font->texture_data, 0, 2 * initial_texture_size * initial_texture_size);
  for(ch = 32; ch < 127; ch++) {   // render ASCII glyphs
    if(render_glyph(font, ch) == 0) { FTGL_Destroy(font); return NULL; }
  }
  save_texture(font);
  return font;
}

void FTGL_Destroy(FTGL_Font* font) {
  if(!font)
    return;
  free((void *)font->glyph);
  free((void *)font->texture_data);
  memset(font, 0, sizeof(*font));
  free((void *)font);
}

long FTGL_Prepare(FTGL_Font* font, const char* str) {
  long i, j, ucs;
  long length = 0;
  int changed = 0, gid, prev_gid = 0;
  if(!font->glyph || !font->texture_data) return -1;
  for(i = 0; str[i]; i++) {
    if((((const unsigned char*)str)[i] & 0xC0) == 0x80)  // incomplete part of an UTF-8 character
      continue;
    if(str[i] > 0 && str[i] < 127)   // standard ASCII; rendered in FTGL_Create
      gid = str[i] - 32;
    else {
      gid = -1;
      ucs = utf8_to_ucs(&str[i]);
      for(j = 95; j < font->num_glyphs; j++)
        if(font->glyph[j].character == ucs)
          gid = j;
      if(gid == -1) {
        if(render_glyph(font, ucs) == 0) return -1;
        changed = 1;
        if(font->glyph[font->num_glyphs-1].character == ucs)
          gid = font->num_glyphs-1;
      }
    }
    if(gid != -1) {
      length += get_kerning(font, prev_gid, gid); prev_gid = gid;
      length += font->glyph[gid].advance;
    }
  }
  if(changed)
    save_texture(font);   // save new glyphs
  return length;
}

void FTGL_Print(FTGL_Font* font, const char* str, int underline) {
  struct _FTGL_Glyph *glyph;
  int gid, prev_gid = 0;
  int active_underline = 0;
  long i, j, ucs, length = 0;
  
  if(!font->glyph || !font->texture_data) return;
  
  if(font->texture)
    glBindTexture(GL_TEXTURE_2D, font->texture);
  
  if(underline < 1 || underline > 126)
    underline = 0;
  
  for(i = 0; str[i]; i++) {
    if((((const unsigned char *)str)[i] & 0xC0) == 0x80)  // incomplete part of an UTF-8 character
      continue;
    
    if(str[i] == underline && underline > 1) {
      active_underline = 1;
      continue;
    }
    
    if(str[i] > 0 && str[i] < 127)
      gid = str[i] - 32;
    else {
      gid = -1;
      ucs = utf8_to_ucs(&str[i]);
      for(j = 95; j < font->num_glyphs; j++)
        if(font->glyph[j].character == ucs)
          gid = j;
      if(gid == -1)   // not found (this only happens if FTGL_Prepare wasn't used)
        continue;
    }
    
    if(font->use_kerning) {
      j = get_kerning(font, prev_gid, gid);
      length += j;
      glTranslated(j, 0, 0);
    }
    prev_gid = gid;
    
    glyph = &font->glyph[gid];
#if 1
    if(glyph->w != 0) {
      glTranslated(glyph->offleft, glyph->offtop, 0);
      glBegin(GL_QUADS);
        glTexCoord2d(glyph->l, glyph->d); glVertex2d(0, 0);
        glTexCoord2d(glyph->r, glyph->d); glVertex2d(glyph->w, 0);
        glTexCoord2d(glyph->r, glyph->u); glVertex2d(glyph->w, glyph->h);
        glTexCoord2d(glyph->l, glyph->u); glVertex2d(0, glyph->h);
      glEnd();
      glTranslated(-glyph->offleft, -glyph->offtop, 0);
    }
    if(active_underline && glyph->advance != 0)
      draw_underline(font, 0, glyph->advance);
    active_underline = 0;
    glTranslated(glyph->advance, 0, 0);
    length += glyph->advance;
#else
#define SC (1.0/font->height)
    if(glyph->w != 0) {
      glTranslated(glyph->offleft*SC, glyph->offtop*SC, 0);
      glBegin(GL_QUADS);
        glTexCoord2d(glyph->l, glyph->d); glVertex2d(0, 0);
        glTexCoord2d(glyph->r, glyph->d); glVertex2d(glyph->w*SC, 0);
        glTexCoord2d(glyph->r, glyph->u); glVertex2d(glyph->w*SC, glyph->h*SC);
        glTexCoord2d(glyph->l, glyph->u); glVertex2d(0, glyph->h*SC);
      glEnd();
      glTranslated(-glyph->offleft*SC, -glyph->offtop*SC, 0);
    }
    if(active_underline && glyph->advance != 0)
      draw_underline(font, 0, glyph->advance*SC);
    active_underline = 0;
    glTranslated(glyph->advance*SC, 0, 0);
    length += glyph->advance*SC;
#endif
  }
  if(underline == 1)
    draw_underline(font, -length, 0);
}

